from .lazydeepClassifier import LazyDeepClassifier, LazyClassifier
from .lazydeepRegressor import LazyDeepRegressor, LazyRegressor
from .lazydeepMTS import LazyDeepMTS, LazyMTS

__all__ = [
    "LazyClassifier",
    "LazyRegressor",
    "LazyDeepClassifier",
    "LazyDeepRegressor",
    "LazyMTS",
    "LazyDeepMTS",
]



# Author: @thierrymoudiki
# Date: 2025-01-09

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import scipy

from collections import namedtuple
from sklearn.base import BaseEstimator, ClassifierMixin
from scipy.optimize import differential_evolution
from dataclasses import dataclass
from functools import partial
from typing import Dict, Any, List
from sklearn.base import clone
from sklearn.base import BaseEstimator, ClassifierMixin
from scipy.optimize import minimize, differential_evolution
from sklearn.ensemble import RandomForestRegressor
from sklearn.kernel_ridge import KernelRidge
from sklearn.linear_model import ElasticNet
from sklearn.datasets import load_diabetes, fetch_california_housing
from sklearn.model_selection import train_test_split
from sklearn.metrics import mean_squared_error
from warnings import filterwarnings

from .quantileregression import QuantileRegressor
from nnetsauce.multitask.simplemultitaskClassifier import SimpleMultitaskClassifier

filterwarnings("ignore")


@dataclass
class QuantileResult:
    params: Dict[str, Any]
    predictions: np.ndarray
    loss: float
    iterations: int


class QuantileClassifier(BaseEstimator, ClassifierMixin):
    """
    Quantile Classifier.

    Parameters:

        obj: base model (classification model)
            The base classifier from which to build a
            quantile classifier.

        level: int, default=95
            The level of the quantiles to compute.

        scoring: str, default="predictions"
            The scoring to use for the optimization and constructing
            prediction intervals (predictions, residuals, conformal,
              studentized, conformal-studentized).

    Attributes:

        obj_ : base model (classification model)
            The base classifier from which to build a
            quantile classifier.

        offset_multipliers_ : list
            The multipliers for the offset.

        scoring_residuals_ : list
            The residuals for the scoring.

        student_multiplier_ : float
            The multiplier for the student.


    """

    def __init__(self, obj, level=95, scoring="predictions"):
        assert scoring in (
            "predictions",
            "residuals",
            "conformal",
            "studentized",
            "conformal-studentized",
        ), "scoring must be 'predictions' or 'residuals'"
        self.obj = obj
        quantileregressor = QuantileRegressor(self.obj)
        quantileregressor.predict = partial(quantileregressor.predict, return_pi=False)
        self.obj_ = SimpleMultitaskClassifier(quantileregressor)

    def fit(self, X, y, **kwargs):
        self.obj_.fit(X, y, **kwargs)

    def predict(self, X, **kwargs):
        return self.obj_.predict(X, **kwargs)

    def predict_proba(self, X, **kwargs):
        return self.obj_.predict_proba(X, **kwargs)



import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin
from nnetsauce.multitask.simplemultitaskClassifier import (
    SimpleMultitaskClassifier,
)
from nnetsauce.neuralnet import NeuralNetRegressor


class NeuralNetClassifier(BaseEstimator, ClassifierMixin):
    """
    (Pretrained) Neural Network Classifier.

    Parameters:

        hidden_layer_sizes : tuple, default=(100,)
            The number of neurons in each hidden layer.
        max_iter : int, default=100
            The maximum number of iterations to train the model.
        learning_rate : float, default=0.01
            The learning rate for the optimizer.
        l1_ratio : float, default=0.5
            The ratio of L1 regularization.
        alpha : float, default=1e-6
            The regularization parameter.
        activation_name : str, default="relu"
            The activation function to use.
        dropout : float, default=0.0
            The dropout rate.
        random_state : int, default=None
            The random state for the random number generator.
        weights : list, default=None
            The weights to initialize the model with.

    Attributes:

        weights : list
            The weights of the model.
        params : list
            The parameters of the model.
        scaler_ : sklearn.preprocessing.StandardScaler
            The scaler used to standardize the input features.
        y_mean_ : float
            The mean of the target variable.

    Methods:

        fit(X, y)
            Fit the model to the data.
        predict(X)
            Predict the target variable.
        predict_proba(X)
            Predict the probability of the target variable.
        get_weights()
            Get the weights of the model.
        set_weights(weights)
            Set the weights of the model.
    """

    def __init__(
        self,
        hidden_layer_sizes=(100,),
        max_iter=100,
        learning_rate=0.01,
        weights=None,
        l1_ratio=0.5,
        alpha=1e-6,
        activation_name="relu",
        dropout=0.0,
        random_state=None,
    ):
        self.hidden_layer_sizes = hidden_layer_sizes
        self.max_iter = max_iter
        self.learning_rate = learning_rate
        self.weights = weights
        self.l1_ratio = l1_ratio
        self.alpha = alpha
        self.activation_name = activation_name
        self.dropout = dropout
        self.random_state = random_state
        self.regr = None

    def fit(self, X, y):
        """Fit the model to the data.

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number of samples and
                n_features is the number of features.
            y: array-like, shape = [n_samples]
                Target values.
        """
        regressor = NeuralNetRegressor(
            hidden_layer_sizes=self.hidden_layer_sizes,
            max_iter=self.max_iter,
            learning_rate=self.learning_rate,
            weights=self.weights,
            l1_ratio=self.l1_ratio,
            alpha=self.alpha,
            activation_name=self.activation_name,
            dropout=self.dropout,
            random_state=self.random_state,
        )
        self.regr = SimpleMultitaskClassifier(regressor)
        self.regr.fit(X, y)
        self.classes_ = np.unique(y)
        self.n_classes_ = len(self.classes_)
        self.n_tasks_ = 1
        self.n_features_in_ = X.shape[1]
        self.n_outputs_ = 1
        self.n_samples_fit_ = X.shape[0]
        self.n_samples_test_ = X.shape[0]
        self.n_features_out_ = 1
        self.n_outputs_ = 1
        self.n_features_in_ = X.shape[1]
        self.n_features_out_ = 1
        self.n_outputs_ = 1
        return self

    def predict_proba(self, X):
        """Predict the probability of the target variable.

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number of samples and
                n_features is the number of features.
        """
        return self.regr.predict_proba(X)

    def predict(self, X):
        """Predict the target variable.

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number of samples and
                n_features is the number of features.
        """
        return self.regr.predict(X)


"""Progbar class."""

# Authors: Thierry Moudiki
#
# License: BSD 3


import time
import numpy as np
import sys
import collections


# progress bar adapted from Keras' training generator
class Progbar(object):
    """Displays a progress bar.

    # Arguments
        target: Total number of steps expected, None if unknown.
        width: Progress bar width on screen.
        verbose: Verbosity mode, 0 (silent), 1 (verbose), 2 (semi-verbose)
        stateful_metrics: Iterable of string names of metrics that
            should *not* be averaged over time. Metrics in this list
            will be displayed as-is. All others will be averaged
            by the progbar before display.
        interval: Minimum visual progress update interval (in seconds).
    """

    def __init__(
        self, target, width=30, verbose=1, interval=0.05, stateful_metrics=None
    ):
        self.target = target
        self.width = width
        self.verbose = verbose
        self.interval = interval
        if stateful_metrics:
            self.stateful_metrics = set(stateful_metrics)
        else:
            self.stateful_metrics = set()

        self._dynamic_display = (
            hasattr(sys.stdout, "isatty") and sys.stdout.isatty()
        ) or "ipykernel" in sys.modules
        self._total_width = 0
        self._seen_so_far = 0
        self._values = collections.OrderedDict()
        self._start = time.time()
        self._last_update = 0

    def update(self, current, values=None):
        """Updates the progress bar.

        # Arguments
            current: Index of current step.
            values: List of tuples:
                `(name, value_for_last_step)`.
                If `name` is in `stateful_metrics`,
                `value_for_last_step` will be displayed as-is.
                Else, an average of the metric over time will be displayed.
        """
        values = values or []
        for k, v in values:
            if k not in self.stateful_metrics:
                if k not in self._values:
                    self._values[k] = [
                        v * (current - self._seen_so_far),
                        current - self._seen_so_far,
                    ]
                else:
                    self._values[k][0] += v * (current - self._seen_so_far)
                    self._values[k][1] += current - self._seen_so_far
            else:
                # Stateful metrics output a numeric value.  This representation
                # means "take an average from a single value" but keeps the
                # numeric formatting.
                self._values[k] = [v, 1]
        self._seen_so_far = current

        now = time.time()
        info = " - %.0fs" % (now - self._start)
        if self.verbose == 1:
            if (
                now - self._last_update < self.interval
                and self.target is not None
                and current < self.target
            ):
                return

            prev_total_width = self._total_width
            if self._dynamic_display:
                sys.stdout.write("\b" * prev_total_width)
                sys.stdout.write("\r")
            else:
                sys.stdout.write("\n")

            if self.target is not None:
                numdigits = int(np.floor(np.log10(self.target))) + 1
                barstr = "%%%dd/%d [" % (numdigits, self.target)
                bar = barstr % current
                prog = float(current) / self.target
                prog_width = int(self.width * prog)
                if prog_width > 0:
                    # bar += ('=' * (prog_width - 1))
                    bar += "█" * (prog_width - 1)
                    if current < self.target:
                        # bar += '>'
                        bar += "█"
                    else:
                        # bar += '='
                        bar += "█"
                bar += "." * (self.width - prog_width)
                bar += "]"
            else:
                bar = "%7d/Unknown" % current

            self._total_width = len(bar)
            sys.stdout.write(bar)

            if current:
                time_per_unit = (now - self._start) / current
            else:
                time_per_unit = 0
            if self.target is not None and current < self.target:
                eta = time_per_unit * (self.target - current)
                if eta > 3600:
                    eta_format = "%d:%02d:%02d" % (
                        eta // 3600,
                        (eta % 3600) // 60,
                        eta % 60,
                    )
                elif eta > 60:
                    eta_format = "%d:%02d" % (eta // 60, eta % 60)
                else:
                    eta_format = "%ds" % eta

                info = " - ETA: %s" % eta_format
            else:
                if time_per_unit >= 1:
                    info += " %.0fs/step" % time_per_unit
                elif time_per_unit >= 1e-3:
                    info += " %.0fms/step" % (time_per_unit * 1e3)
                else:
                    info += " %.0fus/step" % (time_per_unit * 1e6)

            for k in self._values:
                info += " - %s:" % k
                if isinstance(self._values[k], list):
                    avg = np.mean(self._values[k][0] / max(1, self._values[k][1]))
                    if abs(avg) > 1e-3:
                        info += " %.4f" % avg
                    else:
                        info += " %.4e" % avg
                else:
                    info += " %s" % self._values[k]

            self._total_width += len(info)
            if prev_total_width > self._total_width:
                info += " " * (prev_total_width - self._total_width)

            if self.target is not None and current >= self.target:
                info += "\n"

            sys.stdout.write(info)
            sys.stdout.flush()

        elif self.verbose == 2:
            if self.target is None or current >= self.target:
                for k in self._values:
                    info += " - %s:" % k
                    avg = np.mean(self._values[k][0] / max(1, self._values[k][1]))
                    if avg > 1e-3:
                        info += " %.4f" % avg
                    else:
                        info += " %.4e" % avg
                info += "\n"

                sys.stdout.write(info)
                sys.stdout.flush()

        self._last_update = now

    def add(self, n, values=None):
        self.update(self._seen_so_far + n, values)



import numpy as np
import pandas as pd


def getsims(input_data, ix):

    if isinstance(input_data[0], pd.DataFrame):  # kde
        n_sims = len(input_data)
        res = [input_data[i].iloc[:, ix].values for i in range(n_sims)]
        return np.asarray(res).T

    if isinstance(input_data[0], tuple):  # GP posterior
        h = len(input_data)
        n_sims = len(input_data[0][0][0])
        res = [[input_data[hx][ix][0][i] for i in range(n_sims)] for hx in range(h)]
        return np.asarray(res)


def getsimsxreg(sims_ix, output_dates, target_cols):
    """Get simulations from indices when using external regressors"""
    # Convert numpy array to list of DataFrames
    sims = tuple(
        pd.DataFrame(
            sims_ix[:, [i]],  # Keep 2D array shape with single column
            columns=[target_cols[i]],  # Use target column name
            index=output_dates,
        )
        for i in range(sims_ix.shape[1])  # Iterate over columns
    )

    return sims



try:
    import jax
    import jax.numpy as jnp
    from jax import grad, jit, vmap
except ImportError:
    pass

import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.preprocessing import StandardScaler
from functools import partial
from ..utils import matrixops as mo


def predict_internal(params, inputs, activation_func="relu", dropout=0.0, seed=123):
    # Ensure inputs is 2D with correct shape
    inputs = jnp.asarray(inputs)
    if len(inputs.shape) == 1:
        inputs = inputs.reshape(1, -1)
    elif len(inputs.shape) == 2 and inputs.shape[0] == 1:
        inputs = inputs.T
    if activation_func == "tanh":
        for W, b in params[:-1]:  # All layers except last
            outputs = jnp.dot(inputs, W) + b
            inputs = mo.dropout(x=jnp.tanh(outputs), drop_prob=dropout, seed=seed)
    elif activation_func == "relu":
        for W, b in params[:-1]:
            outputs = jnp.dot(inputs, W) + b
            inputs = mo.dropout(x=jnp.maximum(0, outputs), drop_prob=dropout, seed=seed)
    elif activation_func == "sigmoid":
        for W, b in params[:-1]:
            outputs = jnp.dot(inputs, W) + b
            inputs = mo.dropout(x=jax.nn.sigmoid(outputs), drop_prob=dropout, seed=seed)
    elif activation_func == "linear":
        for W, b in params[:-1]:
            outputs = jnp.dot(inputs, W) + b
            inputs = mo.dropout(x=outputs, drop_prob=dropout, seed=seed)
    else:
        raise ValueError(f"Activation function {activation_func} not supported")

    # Final layer (no activation)
    W, b = params[-1]
    outputs = jnp.dot(inputs, W) + b
    return mo.dropout(x=outputs.reshape(-1, 1), drop_prob=dropout, seed=seed)


def loss(
    params,
    inputs,
    targets,
    l1_ratio=0.5,
    alpha=0.01,
    activation_func="relu",
    dropout=0.0,
    seed=123,
):
    preds = predict_internal(
        params,
        inputs,
        activation_func=activation_func,
        dropout=dropout,
        seed=seed,
    )
    mse = jnp.sum((preds - targets) ** 2)

    # Calculate L1 and L2 regularization for all parameters
    l1_penalty = sum(jnp.sum(jnp.abs(W)) + jnp.sum(jnp.abs(b)) for W, b in params)
    l2_penalty = sum(jnp.sum(jnp.square(W)) + jnp.sum(jnp.square(b)) for W, b in params)

    # Combine MSE with elastic net regularization
    res = mse + alpha * (l1_ratio * l1_penalty + (1 - l1_ratio) * l2_penalty)
    return res


def initialize_params(input_dim, hidden_layer_sizes, random_state=None):
    """Initialize network parameters."""
    if random_state is not None:
        rng = jax.random.PRNGKey(random_state)
    else:
        rng = jax.random.PRNGKey(0)

    # Layer dimensions including input and output
    layer_sizes = [input_dim] + list(hidden_layer_sizes) + [1]

    # Initialize parameters for each layer
    params = []
    for i in range(len(layer_sizes) - 1):
        rng, key = jax.random.split(rng)
        # Xavier/Glorot initialization
        scale = jnp.sqrt(2.0 / (layer_sizes[i] + layer_sizes[i + 1]))
        W = scale * jax.random.normal(key, (layer_sizes[i], layer_sizes[i + 1]))
        b = jnp.zeros(layer_sizes[i + 1])
        params.append((W, b))
    return params


class NeuralNetRegressor(BaseEstimator, RegressorMixin):
    """
    (Pretrained) Neural Network Regressor.

    Parameters:

        hidden_layer_sizes : tuple, default=(100,)
            The number of neurons in each hidden layer.
        max_iter : int, default=100
            The maximum number of iterations to train the model.
        learning_rate : float, default=0.01
            The learning rate for the optimizer.
        l1_ratio : float, default=0.5
            The ratio of L1 regularization.
        alpha : float, default=1e-6
            The regularization parameter.
        activation_name : str, default="relu"
            The activation function to use.
        dropout : float, default=0.0
            The dropout rate.
        random_state : int, default=None
            The random state for the random number generator.
        weights : list, default=None
            The weights to initialize the model with.

    Attributes:

        weights : list
            The weights of the model.
        params : list
            The parameters of the model.
        scaler_ : sklearn.preprocessing.StandardScaler
            The scaler used to standardize the input features.
        y_mean_ : float
            The mean of the target variable.

    Methods:

        fit(X, y)
            Fit the model to the data.
        predict(X)
            Predict the target variable.
        get_weights()
            Get the weights of the model.
        set_weights(weights)
            Set the weights of the model.
    """

    def __init__(
        self,
        hidden_layer_sizes=None,
        max_iter=100,
        learning_rate=0.01,
        l1_ratio=0.5,
        alpha=1e-6,
        activation_name="relu",
        dropout=0,
        weights=None,
        random_state=None,
    ):
        if weights is None and hidden_layer_sizes is None:
            hidden_layer_sizes = (100,)  # default value if neither is provided
        self.hidden_layer_sizes = hidden_layer_sizes
        self.max_iter = max_iter
        self.learning_rate = learning_rate
        self.l1_ratio = l1_ratio
        self.alpha = alpha
        self.activation_name = activation_name
        self.dropout = dropout
        self.weights = weights
        self.random_state = random_state
        self.params = None
        self.scaler_ = StandardScaler()
        self.y_mean_ = None

    def _validate_weights(self, input_dim):
        """Validate that weights dimensions are coherent."""
        if not self.weights:
            return False

        try:
            # Check each layer's weights and biases
            prev_dim = input_dim
            for W, b in self.weights:
                # Check weight matrix dimensions
                if W.shape[0] != prev_dim:
                    raise ValueError(
                        f"Weight matrix input dimension {W.shape[0]} does not match, previous layer output dimension {prev_dim}"
                    )
                # Check bias dimension matches weight matrix output
                if W.shape[1] != b.shape[0]:
                    raise ValueError(
                        f"Bias dimension {b.shape[0]} does not match weight matrix, output dimension {W.shape[1]}"
                    )
                prev_dim = W.shape[1]

            # Check final output dimension is 1 for regression
            if prev_dim != 1:
                raise ValueError(
                    f"Final layer output dimension {prev_dim} must be 1 for regression"
                )

            return True
        except (AttributeError, IndexError):
            raise ValueError(
                "Weights format is invalid. Expected list of (weight, bias) tuples"
            )

    def fit(self, X, y):
        # Standardize the input features
        X = self.scaler_.fit_transform(X)
        # Ensure y is 2D for consistency
        y = y.reshape(-1, 1)
        self.y_mean_ = jnp.mean(y)
        y = y - self.y_mean_
        # Validate or initialize weights
        if self.weights is not None:
            if self._validate_weights(X.shape[1]):
                self.params = self.weights
        else:
            if self.hidden_layer_sizes is None:
                raise ValueError(
                    "Either weights or hidden_layer_sizes must be provided"
                )
            self.params = initialize_params(
                X.shape[1], self.hidden_layer_sizes, self.random_state
            )
        loss_fn = partial(loss, l1_ratio=self.l1_ratio, alpha=self.alpha)
        grad_loss = jit(grad(loss_fn))  # compiled gradient evaluation function
        perex_grads = jit(
            vmap(grad_loss, in_axes=(None, 0, 0))
        )  # fast per-example grads
        # Training loop
        for _ in range(self.max_iter):
            grads = perex_grads(self.params, X, y)
            # Average gradients across examples
            grads = jax.tree_map(lambda g: jnp.mean(g, axis=0), grads)
            # Update parameters
            self.params = [
                (W - self.learning_rate * dW, b - self.learning_rate * db)
                for (W, b), (dW, db) in zip(self.params, grads)
            ]
        # Store final weights
        self.weights = self.params
        return self

    def get_weights(self):
        """Return the current weights of the model."""
        if self.weights is None:
            raise ValueError("No weights available. Model has not been fitted yet.")
        return self.weights

    def set_weights(self, weights):
        """Set the weights of the model manually."""
        self.weights = weights
        self.params = weights

    def predict(self, X):
        X = self.scaler_.transform(X)
        if self.params is None:
            raise ValueError("Model has not been fitted yet.")
        predictions = predict_internal(
            self.params,
            X,
            activation_func=self.activation_name,
            dropout=self.dropout,
            seed=self.random_state,
        )
        return predictions.reshape(-1) + self.y_mean_


import numpy as np
from sklearn.ensemble import VotingRegressor


class MedianVotingRegressor(VotingRegressor):
    def predict(self, X):
        """
        Predict using the median of the base regressors' predictions.

        Parameters:
        X (array-like): Feature matrix for predictions.

        Returns:
        y_pred (array): Median of predictions from the base regressors.
        """
        predictions = np.asarray(
            [regressor.predict(X) for regressor in self.estimators_]
        )
        return np.median(predictions, axis=0)



import numpy as np
import ctypes
import os

from six import moves
from .sobol import i4_sobol_generate

# From: https://github.com/PhaethonPrime/hammersley/blob/master/hammersley/sequences.py

# this list of primes allows up to a size 120 vector
saved_primes = [
    2,
    3,
    5,
    7,
    11,
    13,
    17,
    19,
    23,
    29,
    31,
    37,
    41,
    43,
    47,
    53,
    59,
    61,
    67,
    71,
    73,
    79,
    83,
    89,
    97,
    101,
    103,
    107,
    109,
    113,
    127,
    131,
    137,
    139,
    149,
    151,
    157,
    163,
    167,
    173,
    179,
    181,
    191,
    193,
    197,
    199,
    211,
    223,
    227,
    229,
    233,
    239,
    241,
    251,
    257,
    263,
    269,
    271,
    277,
    281,
    283,
    293,
    307,
    311,
    313,
    317,
    331,
    337,
    347,
    349,
    353,
    359,
    367,
    373,
    379,
    383,
    389,
    397,
    401,
    409,
    419,
    421,
    431,
    433,
    439,
    443,
    449,
    457,
    461,
    463,
    467,
    479,
    487,
    491,
    499,
    503,
    509,
    521,
    523,
    541,
    547,
    557,
    563,
    569,
    571,
    577,
    587,
    593,
    599,
    601,
    607,
    613,
    617,
    619,
    631,
    641,
    643,
    647,
    653,
    659,
]


def get_phi(p, k):
    p_ = p
    k_ = k
    phi = 0
    while k_ > 0:
        a = k_ % p
        phi += a / p_
        k_ = int(k_ / p)
        p_ *= p
    return phi


# uniform numbers' generation (python)
def generate_uniform(n_dims=2, n_points=10, seed=123):
    np.random.seed(seed=seed)
    return np.random.random((n_dims, n_points))


# hammersley numbers' generation (python)
def generate_hammersley(
    n_dims=2, n_points=100, primes=None, seed=None
):  # seed, just for 'compatibility'
    def func_hammersley(n_dims=n_dims, n_points=(n_points + 1), primes=primes):
        primes = primes if primes is not None else saved_primes
        for k in moves.range(n_points):
            points = [k / n_points] + [
                get_phi(primes[d], k) for d in moves.range(n_dims - 1)
            ]
            yield points

    return np.array(list(func_hammersley()))[1 : (n_points + 1), :].transpose()


# halton numbers' generation (python)
def generate_halton(
    n_dims=2, n_points=10, primes=None, seed=None
):  # seed, just for 'compatibility'
    def func_halton(n_dims=n_dims, n_points=(n_points + 1), primes=primes):
        primes = primes if primes is not None else saved_primes
        for k in moves.range(n_points):
            points = [get_phi(primes[d], k) for d in moves.range(n_dims)]
            yield points

    return np.array(list(func_halton()))[1 : (n_points + 1), :].transpose()


# sobol numbers' generation (python)
def generate_sobol(n_dims=2, n_points=10, seed=None):  # seed, just for 'compatibility'
    return np.array(i4_sobol_generate(m=n_dims, n=n_points, skip=2))



# Authors: T. Moudiki
#
# License: BSD 3 Clear Clause

import numpy as np
import pandas as pd
from collections import namedtuple
from scipy.stats import norm
from scipy import stats
from statsmodels.tsa.stattools import adfuller
from statsmodels.tsa.stattools import kpss
from ..base import Base
from .mts import MTS
from ..utils import timeseries as ts


class MLARCH(MTS):
    """Machine Learning with ARCH effects for time series forecasting

    Parameters:
            
        model_mean: object of class nnetsauce.MTS
            Model for mean prediction (default: None, uses obj)
            
        model_sigma: object of class nnetsauce.MTS
            Model for residuals volatility prediction (default: None, uses obj)
        
        model_residuals: object of class nnetsauce.MTS
            Model for residuals prediction (default: None, uses obj)
    
    Examples: 

        See examples/mlarch.py
                        
    """
    def __init__(
        self,
        model_mean,
        model_sigma, 
        model_residuals
    ):
        assert isinstance(model_mean, MTS), "model_mean must be an object of class nnetsauce.MTS"
        assert isinstance(model_sigma, MTS), "model_sigma must be an object of class nnetsauce.MTS"
        assert isinstance(model_residuals, MTS), "model_residuals must be an object of class nnetsauce.MTS"
        assert model_sigma.type_pi.startswith("scp") and model_sigma.replications is not None, \
        "for now, the models must be conformalized, i.e type_pi must start with 'scp' and replications must be an integer"
        assert model_residuals.type_pi.startswith("scp") and model_residuals.replications is not None, \
        "for now, the models must be conformalized, i.e type_pi must start with 'scp' and replications must be an integer"        

        self.model_mean = model_mean
        self.model_sigma = model_sigma
        self.model_residuals = model_residuals

        self.mean_residuals_ = None
        self.mean_residuals_wilcoxon_test_ = None
        self.mean_residuals_kpss_test_ = None
        self.standardized_residuals_ = None


    def fit(self, y):
        """Fit the MLARCH model to the time series data.

        Parameters
        ----------
        y : array-like of shape (n_samples,)
            The target time series to be fitted.

        Returns
        -------
        self : object
            Returns self.

        Notes
        -----
        This method:

        1. Fits the mean model to the time series
        2. Performs statistical tests on the residuals (Wilcoxon and KPSS)
        3. Fits the volatility model to the squared residuals
        4. Computes standardized residuals
        5. Fits the residuals model to the standardized residuals
        """
        n = len(y)
        self.model_mean.fit(y.reshape(-1, 1)) 
        # Wilcoxon signed-rank test on residuals (mean = 0)
        self.mean_residuals_wilcoxon_test_ = stats.wilcoxon(self.model_mean.residuals_)
        # KPSS test for stationarity on residuals
        self.mean_residuals_kpss_test_ = kpss(self.model_mean.residuals_, regression='c')
        self.model_sigma.fit(np.log(self.model_mean.residuals_.reshape(-1, 1)**2)) 
        # n//2 here because the model is conformalized
        fitted_sigma = self.model_sigma.residuals_ + np.log(self.model_mean.residuals_**2)[(n//2):,:]
        # standardized residuals
        self.standardized_residuals_ = self.model_mean.residuals_[(n//2):,:]/np.sqrt(np.exp(fitted_sigma))
        self.model_residuals.fit(self.standardized_residuals_.reshape(-1, 1))

        # Calculate AIC
        # Get predictions from all models
        mean_pred = self.model_mean.predict(h=0).values.ravel()
        sigma_pred = self.model_sigma.predict(h=0).values.ravel()
        z_pred = self.model_residuals.predict(h=0).values.ravel()
        
        # Calculate combined predictions
        combined_pred = mean_pred + z_pred * np.sqrt(np.exp(sigma_pred))
        
        # Calculate SSE using the last half of the data (matching standardized_residuals_)
        y_actual = y[(n//2):].ravel()
        self.sse_ = np.sum((y_actual - combined_pred) ** 2)
        
        # Calculate number of parameters (sum of parameters from all three models)
        n_params = (self.model_mean.n_hidden_features + 1 +  # mean model
                   self.model_sigma.n_hidden_features + 1 +  # sigma model
                   self.model_residuals.n_hidden_features + 1)  # residuals model
        
        # Calculate AIC
        n_samples = len(y_actual)
        self.aic_ = n_samples * np.log(self.sse_/n_samples) + 2 * n_params

        return self


    def predict(self, h=5, level=95):
        """Predict (probabilistic) future values of the time series.

        Parameters
        ----------
        h : int, default=5
            The forecast horizon.
        level : int, default=95
            The confidence level for prediction intervals.

        Returns
        -------
        DescribeResult : namedtuple
            A named tuple containing:

            - mean : array-like of shape (h,)
                The mean forecast.
            - sims : array-like of shape (h, n_replications)
                The simulated forecasts.
            - lower : array-like of shape (h,)
                The lower bound of the prediction interval.
            - upper : array-like of shape (h,)
                The upper bound of the prediction interval.

        Notes
        -----
        This method:
        1. Generates mean forecasts using the mean model
        2. Generates standardized residual forecasts using the residuals model
        3. Generates volatility forecasts using the sigma model
        4. Combines these forecasts to generate the final predictions
        5. Computes prediction intervals at the specified confidence level
        """
        DescribeResult = namedtuple(
                "DescribeResult", ("mean", "sims", "lower", "upper")
            )
        mean_forecast = self.model_mean.predict(h=h).values.ravel()
        preds_z = self.model_residuals.predict(h=h)
        preds_sigma = self.model_sigma.predict(h=h)
        sims_z = preds_z.sims
        sims_sigma = preds_sigma.sims 

        f = []
        for i in range(len(sims_z)): 
            f.append(mean_forecast + sims_z[i].values.ravel()*np.sqrt(np.exp(sims_sigma[i].values.ravel())))

        f = np.asarray(f).T
        mean_f = np.mean(f, axis=1)
        alpha = 1 - level/100
        lower_bound = np.quantile(f, alpha/2, axis=1)
        upper_bound = np.quantile(f, 1-alpha/2, axis=1)

        return DescribeResult(mean_f, f, 
                              lower_bound, upper_bound)






# Authors: Thierry Moudiki
#
# License: BSD 3 Clear

import numpy as np
import warnings
from sklearn.base import RegressorMixin
from ..base import Base

try:
    import jax.numpy as jnp
    import jax.nn as jnn
except ImportError:
    pass

from ..utils import matrixops as mo


class RidgeRegressor(Base, RegressorMixin):
    """Basic Ridge Regression model.

    Parameters:
        lambda_: float or array-like
            Ridge regularization parameter(s). Default is 0.
    """

    def __init__(
        self,
        lambda_=0.0,
        n_hidden_features=0,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=0,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        seed=123,
        backend="cpu",
    ):
        super().__init__(
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
            backend=backend,
        )
        self.lambda_ = lambda_
        self.scale_ = {}

    def _center_scale_xy(self, X, y):
        """Center X and y, scale X."""
        n = X.shape[0]

        # Center X and y
        X_mean = np.mean(X, axis=0)
        y_mean = np.mean(y)
        X_centered = X - X_mean
        y_centered = y - y_mean

        # Scale X
        X_scale = np.sqrt(np.sum(X_centered**2, axis=0) / n)
        # Avoid division by zero
        X_scale = np.where(X_scale == 0, 1.0, X_scale)
        X_scaled = X_centered / X_scale

        return X_scaled, y_centered, X_mean, y_mean, X_scale

    def fit(self, X, y):
        """Fit Ridge regression model.

        Parameters:
            X : array-like of shape (n_samples, n_features)
                Training data
            y : array-like of shape (n_samples,)
                Target values

        Returns:
            self : returns an instance of self.
        """
        # Ensure numpy arrays
        X = np.asarray(X)
        y = np.asarray(y)
        print(f"\nInput shapes - X: {X.shape}, y: {y.shape}")
        print(f"First few X values: {X[:2]}")
        print(f"First few y values: {y[:2]}")

        if y.ndim == 2:
            y = y.ravel()

        # Center and scale
        X_scaled, y_centered, self.X_mean_, self.y_mean_, self.X_scale_ = (
            self._center_scale_xy(X, y)
        )

        # SVD decomposition
        U, d, Vt = np.linalg.svd(X_scaled, full_matrices=False)

        # Compute coefficients
        rhs = np.dot(U.T, y_centered)
        d2 = d**2

        print(f"d2 shape: {d2.shape}")
        print(f"rhs shape: {rhs.shape}")
        print(f"Vt shape: {Vt.shape}")

        if np.isscalar(self.lambda_):
            div = d2 + self.lambda_
            a = (d * rhs) / div
            print(f"\nSingle lambda case:")
            print(f"lambda: {self.lambda_}")
            print(f"div shape: {div.shape}")
            print(f"a shape: {a.shape}")
            self.coef_ = np.dot(Vt.T, a) / self.X_scale_
            print(f"coef shape: {self.coef_.shape}")
        else:
            coefs = []
            print(f"\nMultiple lambda case:")
            for lambda_ in self.lambda_:
                print(f"lambda: {lambda_}")
                div = d2 + lambda_
                print(f"div shape: {div.shape}")
                a = (d * rhs) / div
                print(f"a shape: {a.shape}")
                coef = np.dot(Vt.T, a) / self.X_scale_
                print(f"coef shape: {coef.shape}")
                coefs.append(coef)
            self.coef_ = np.array(coefs).T
            print(f"final coefs shape: {self.coef_.shape}")

        # Compute GCV, HKB and LW criteria
        y_pred = self.predict(X)
        try:
            resid = y - y_pred
        except Exception as e:
            resid = y[:, np.newaxis] - y_pred
        n, p = X.shape
        if resid.ndim == 1:
            s2 = np.sum(resid**2) / (n - p)
        else:
            s2 = np.sum(resid**2, axis=0) / (n - p)

        self.HKB_ = (p - 2) * s2 / np.sum(self.coef_**2)
        self.LW_ = (p - 2) * s2 * n / np.sum(y_pred**2)

        if np.isscalar(self.lambda_):
            div = d2 + self.lambda_
            self.GCV_ = np.sum((y - y_pred) ** 2) / (n - np.sum(d2 / div)) ** 2
        else:
            self.GCV_ = []
            for lambda_ in self.lambda_:
                div = d2 + lambda_
                try:
                    gcv = np.sum((y - y_pred) ** 2) / (n - np.sum(d2 / div)) ** 2
                except Exception as e:
                    gcv = (
                        np.sum((y[:, np.newaxis] - y_pred) ** 2)
                        / (n - np.sum(d2 / div)) ** 2
                    )
                self.GCV_.append(gcv)
            self.GCV_ = np.array(self.GCV_)

        return self

    def predict(self, X):
        """Predict using the Ridge regression model.

        Parameters:
            X : array-like of shape (n_samples, n_features)
                Samples to predict for

        Returns:
            y_pred : array-like of shape (n_samples,)
                Returns predicted values.
        """
        # Ensure X is 2D
        X = np.asarray(X)
        if X.ndim == 1:
            X = X.reshape(1, -1)
            
        # Center and scale X
        X_scaled = (X - self.X_mean_) / self.X_scale_

        if self.backend == "cpu":
            if np.isscalar(self.lambda_):
                return (
                    mo.safe_sparse_dot(X_scaled, self.coef_, backend=self.backend)
                    + self.y_mean_
                )
            else:
                return np.array(
                    [
                        mo.safe_sparse_dot(X_scaled, coef, backend=self.backend) + self.y_mean_
                        for coef in self.coef_.T
                    ]
                ).T
        else:
            if np.isscalar(self.lambda_):
                return (
                    mo.safe_sparse_dot(X_scaled, self.coef_, backend=self.backend)
                    + self.y_mean_
                )
            else:
                return jnp.array(
                    [
                        mo.safe_sparse_dot(X_scaled, coef, backend=self.backend) + self.y_mean_
                        for coef in self.coef_.T
                    ]
                ).T

    def decision_function(self, X):
        """Compute the decision function of X.

        Parameters:
            X : array-like of shape (n_samples, n_features)
                Samples

        Returns:
            decision : array-like of shape (n_samples,) or (n_samples, n_lambdas)
                Decision function of the input samples. The order of outputs is the same
                as that of the provided lambda_ values. For a single lambda, returns
                array of shape (n_samples,). For multiple lambdas, returns array of shape
                (n_samples, n_lambdas).
        """
        X = self.cook_test_set(X)

        if self.backend == "cpu":
            if np.isscalar(self.lambda_):
                return mo.safe_sparse_dot(X, self.coef_, backend=self.backend)
            else:
                return np.array(
                    [
                        mo.safe_sparse_dot(X, coef, backend=self.backend)
                        for coef in self.coef_.T
                    ]
                ).T
        else:
            if np.isscalar(self.lambda_):
                return mo.safe_sparse_dot(X, self.coef_, backend=self.backend)
            else:
                return jnp.array(
                    [
                        mo.safe_sparse_dot(X, coef, backend=self.backend)
                        for coef in self.coef_.T
                    ]
                ).T



# Author: @thierrymoudiki
# Date: 2025-01-03

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import scipy

from collections import namedtuple
from sklearn.base import BaseEstimator, RegressorMixin
from scipy.optimize import differential_evolution
from dataclasses import dataclass
from typing import Dict, Any, List
from sklearn.base import clone
from sklearn.base import BaseEstimator, RegressorMixin
from scipy.optimize import minimize, differential_evolution
from sklearn.ensemble import RandomForestRegressor
from sklearn.kernel_ridge import KernelRidge
from sklearn.linear_model import ElasticNet
from sklearn.datasets import load_diabetes, fetch_california_housing
from sklearn.model_selection import train_test_split
from sklearn.metrics import mean_squared_error
from warnings import filterwarnings

filterwarnings("ignore")


@dataclass
class QuantileResult:
    params: Dict[str, Any]
    predictions: np.ndarray
    loss: float
    iterations: int


class QuantileRegressor(BaseEstimator, RegressorMixin):
    """
    Quantile Regressor.

    Parameters:

        obj: base model (regression model)
            The base regressor from which to build a
            quantile regressor.

        level: int, default=95
            The level of the quantiles to compute.

        scoring: str, default="predictions"
            The scoring to use for the optimization and constructing
            prediction intervals (predictions, residuals, conformal,
              studentized, conformal-studentized).

    Attributes:

        obj_ : base model (regression model)
            The base regressor from which to build a
            quantile regressor.

        offset_multipliers_ : list
            The multipliers for the offset.

        scoring_residuals_ : list
            The residuals for the scoring.

        student_multiplier_ : float
            The multiplier for the student.


    """

    def __init__(self, obj, level=95, scoring="predictions"):
        assert scoring in (
            "predictions",
            "residuals",
            "conformal",
            "studentized",
            "conformal-studentized",
        ), "scoring must be 'predictions' or 'residuals'"
        self.obj = obj
        low_risk_level = (1 - level / 100) / 2
        self.quantiles = [low_risk_level, 0.5, 1 - low_risk_level]
        self.scoring = scoring
        self.offset_multipliers_ = None
        self.obj_ = None
        self.scoring_residuals_ = None
        self.student_multiplier_ = None

    def _compute_quantile_loss(self, residuals: np.ndarray, quantile: float) -> float:
        """
        Compute the quantile loss for a given set of residuals and quantile.
        """
        if not 0 < quantile < 1:
            raise ValueError("Quantile should be between 0 and 1.")
        loss = quantile * (residuals >= 0) + (quantile - 1) * (residuals < 0)
        return np.mean(residuals * loss)

    def _optimize_multiplier(
        self,
        y: np.ndarray,
        base_predictions: np.ndarray,
        prev_predictions: np.ndarray = None,
        scoring_residuals: np.ndarray = None,
        quantile: float = 0.5,
    ) -> float:
        """
        Optimize the multiplier for a given quantile.
        """
        if not 0 < quantile < 1:
            raise ValueError("Quantile should be between 0 and 1.")

        def objective(log_multiplier):
            """
            Objective function for optimization.
            """
            # Convert to positive multiplier using exp
            multiplier = np.exp(log_multiplier[0])
            if self.scoring == "predictions":
                assert base_predictions is not None, "base_predictions must be not None"
                # Calculate predictions
                if prev_predictions is None:
                    # For first quantile, subtract from conditional expectation
                    predictions = base_predictions - multiplier * np.abs(
                        base_predictions
                    )
                else:
                    # For other quantiles, add to previous quantile
                    offset = multiplier * np.abs(base_predictions)
                    predictions = prev_predictions + offset
            elif self.scoring in ("residuals", "conformal"):
                assert (
                    scoring_residuals is not None
                ), "scoring_residuals must be not None"
                # print("scoring_residuals", scoring_residuals)
                # Calculate predictions
                if prev_predictions is None:
                    # For first quantile, subtract from conditional expectation
                    predictions = base_predictions - multiplier * np.std(
                        scoring_residuals
                    )
                    # print("predictions", predictions)
                else:
                    # For other quantiles, add to previous quantile
                    offset = multiplier * np.std(scoring_residuals)
                    predictions = prev_predictions + offset
            elif self.scoring in ("studentized", "conformal-studentized"):
                assert (
                    scoring_residuals is not None
                ), "scoring_residuals must be not None"
                # Calculate predictions
                if prev_predictions is None:
                    # For first quantile, subtract from conditional expectation
                    predictions = (
                        base_predictions - multiplier * self.student_multiplier_
                    )
                    # print("predictions", predictions)
                else:
                    # For other quantiles, add to previous quantile
                    offset = multiplier * self.student_multiplier_
                    predictions = prev_predictions + offset
            else:
                raise ValueError("Invalid argument 'scoring'")

            residuals = y - predictions
            return self._compute_quantile_loss(residuals, quantile)

        # Optimize in log space for numerical stability
        # bounds = [(-10, 10)]  # log space bounds
        bounds = [(-100, 100)]  # log space bounds
        result = differential_evolution(
            objective,
            bounds,
            # popsize=15,
            # maxiter=100,
            # tol=1e-4,
            popsize=25,
            maxiter=200,
            tol=1e-6,
            disp=False,
        )

        return np.exp(result.x[0])

    def fit(self, X: np.ndarray, y: np.ndarray):
        """Fit the model to the data.

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number of samples and
                n_features is the number of features.
            y: array-like, shape = [n_samples]
                Target values.
        """
        self.obj_ = clone(self.obj)
        if self.scoring in ("predictions", "residuals"):
            self.obj_.fit(X, y)
            base_predictions = self.obj_.predict(X)
            scoring_residuals = y - base_predictions
            self.scoring_residuals_ = scoring_residuals
        elif self.scoring == "conformal":
            X_train, X_calib, y_train, y_calib = train_test_split(
                X, y, test_size=0.5, random_state=42
            )
            self.obj_.fit(X_train, y_train)
            scoring_residuals = y_calib - self.obj_.predict(
                X_calib
            )  # These are calibration predictions
            self.scoring_residuals_ = scoring_residuals
            # Update base_predictions to use training predictions for optimization
            self.obj_.fit(X_calib, y_calib)
            base_predictions = self.obj_.predict(X_calib)
        elif self.scoring in ("studentized", "conformal-studentized"):
            # Calculate student multiplier
            if self.scoring == "conformal-studentized":
                X_train, X_calib, y_train, y_calib = train_test_split(
                    X, y, test_size=0.5, random_state=42
                )
                self.obj_.fit(X_train, y_train)
                scoring_residuals = y_calib - self.obj_.predict(X_calib)
                # Calculate studentized multiplier using calibration data
                self.student_multiplier_ = np.std(y_calib, ddof=1) / np.sqrt(
                    len(y_calib)
                )
                self.obj_.fit(X_calib, y_calib)
                base_predictions = self.obj_.predict(X_calib)
            else:  # regular studentized
                self.obj_.fit(X, y)
                base_predictions = self.obj_.predict(X)
                scoring_residuals = y - base_predictions
                self.student_multiplier_ = np.std(y, ddof=1) / np.sqrt(len(y))

        # Initialize storage for multipliers
        self.offset_multipliers_ = []
        # Keep track of current predictions for each quantile
        current_predictions = None

        # Fit each quantile sequentially
        for i, quantile in enumerate(self.quantiles):
            if self.scoring == "predictions":
                multiplier = self._optimize_multiplier(
                    y=y,
                    base_predictions=base_predictions,
                    prev_predictions=current_predictions,
                    quantile=quantile,
                )

                self.offset_multipliers_.append(multiplier)

                # Update current predictions
                if current_predictions is None:
                    # First quantile (lowest)
                    current_predictions = base_predictions - multiplier * np.abs(
                        base_predictions
                    )
                else:
                    # Subsequent quantiles
                    offset = multiplier * np.abs(base_predictions)
                    current_predictions = current_predictions + offset
            elif self.scoring == "residuals":
                multiplier = self._optimize_multiplier(
                    y=y,
                    base_predictions=base_predictions,
                    scoring_residuals=scoring_residuals,
                    prev_predictions=current_predictions,
                    quantile=quantile,
                )

                self.offset_multipliers_.append(multiplier)

                # Update current predictions
                if current_predictions is None:
                    # First quantile (lowest)
                    current_predictions = base_predictions - multiplier * np.std(
                        scoring_residuals
                    )
                else:
                    # Subsequent quantiles
                    offset = multiplier * np.std(scoring_residuals)
                    current_predictions = current_predictions + offset
            elif self.scoring == "conformal":
                multiplier = self._optimize_multiplier(
                    y=y_calib,
                    base_predictions=base_predictions,
                    scoring_residuals=scoring_residuals,
                    prev_predictions=current_predictions,
                    quantile=quantile,
                )

                self.offset_multipliers_.append(multiplier)

                # Update current predictions
                if current_predictions is None:
                    # First quantile (lowest)
                    current_predictions = base_predictions - multiplier * np.std(
                        scoring_residuals
                    )
                else:
                    # Subsequent quantiles
                    offset = multiplier * np.std(scoring_residuals)
                    current_predictions = current_predictions + offset
            elif self.scoring in ("studentized", "conformal-studentized"):
                multiplier = self._optimize_multiplier(
                    y=y_calib if self.scoring == "conformal-studentized" else y,
                    base_predictions=base_predictions,
                    scoring_residuals=scoring_residuals,
                    prev_predictions=current_predictions,
                    quantile=quantile,
                )

                self.offset_multipliers_.append(multiplier)

                # Update current predictions
                if current_predictions is None:
                    current_predictions = (
                        base_predictions - multiplier * self.student_multiplier_
                    )
                else:
                    offset = multiplier * self.student_multiplier_
                    current_predictions = current_predictions + offset

        return self

    def predict(self, X, return_pi=False):
        """Predict the target variable.

        Parameters:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number of samples and
                n_features is the number of features.

            return_pi: bool, default=True
                Whether to return the prediction intervals.
        """
        if self.obj_ is None or self.offset_multipliers_ is None:
            raise ValueError("Model not fitted yet.")

        base_predictions = self.obj_.predict(X)
        all_predictions = []

        if self.scoring == "predictions":

            # Generate first quantile
            current_predictions = base_predictions - self.offset_multipliers_[
                0
            ] * np.abs(base_predictions)
            all_predictions.append(current_predictions)

            # Generate remaining quantiles
            for multiplier in self.offset_multipliers_[1:]:
                offset = multiplier * np.abs(base_predictions)
                current_predictions = current_predictions + offset
                all_predictions.append(current_predictions)

        elif self.scoring in ("residuals", "conformal"):

            # Generate first quantile
            current_predictions = base_predictions - self.offset_multipliers_[
                0
            ] * np.std(self.scoring_residuals_)
            all_predictions.append(current_predictions)

            # Generate remaining quantiles
            for multiplier in self.offset_multipliers_[1:]:
                offset = multiplier * np.std(self.scoring_residuals_)
                current_predictions = current_predictions + offset
                all_predictions.append(current_predictions)

        elif self.scoring in ("studentized", "conformal-studentized"):
            # Generate first quantile
            current_predictions = (
                base_predictions
                - self.offset_multipliers_[0] * self.student_multiplier_
            )
            all_predictions.append(current_predictions)

            # Generate remaining quantiles
            for multiplier in self.offset_multipliers_[1:]:
                offset = multiplier * self.student_multiplier_
                current_predictions = current_predictions + offset
                all_predictions.append(current_predictions)

        if return_pi == False:
            return np.asarray(all_predictions[1])

        DescribeResult = namedtuple(
            "DecribeResult", ["mean", "lower", "upper", "median"]
        )
        DescribeResult.mean = base_predictions
        DescribeResult.lower = np.asarray(all_predictions[0])
        DescribeResult.median = np.asarray(all_predictions[1])
        DescribeResult.upper = np.asarray(all_predictions[2])
        return DescribeResult



import numpy as np
import pandas as pd
from copy import deepcopy
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_is_fitted
from ..base import Base
from ..custom import CustomRegressor
from ..deep import DeepRegressor


def _update_mean(mean_old, n_obs, new_vectors):
    return (n_obs * mean_old + new_vectors) / (n_obs + 1)


class RegressorUpdater(BaseEstimator, RegressorMixin):
    """
    Update a regression model with new observations

    Parameters
    ----------
    regr: object
        A regression model with a coef_ attribute
    alpha: float
        Updating factor's exponent

    Attributes
    ----------
    n_obs_: int
        Number of observations
    coef_: np.ndarray
        Coefficients of the model
    updating_factor_: float
        Updating factor

    """

    def __init__(self, regr, alpha=0.5):
        self.regr = regr
        self.alpha = alpha
        self.n_obs_ = None
        self.coef_ = None
        self.updating_factor_ = None
        try:
            self.coef_ = self.regr.coef_
            if isinstance(self.regr, Base):
                self.n_obs_ = self.regr.scaler_.n_samples_seen_
        except AttributeError:
            pass

    def fit(self, X, y, **kwargs):

        if isinstance(self.regr, CustomRegressor):  # nnetsauce model not deep ---
            if check_is_fitted(self.regr) == False:
                self.regr.fit(X, y, **kwargs)
                self.n_obs_ = X.shape[0]
                if hasattr(self.regr, "coef_"):
                    self.coef_ = self.regr.coef_
                return self
            self.n_obs_ = self.regr.scaler_.n_samples_seen_
            if hasattr(self.regr, "coef_"):
                self.coef_ = self.regr.coef_
            return self

        if (
            hasattr(self.regr, "coef_") == False
        ):  # sklearn model or CustomRegressor model ---
            self.regr.fit(X, y)
            self.n_obs_ = X.shape[0]
            self.regr.fit(X, y)
            if hasattr(self.regr, "stacked_obj"):
                self.coef_ = self.regr.stacked_obj.coef_
            else:
                self.coef_ = self.regr.coef_
            return self
        self.n_obs_ = X.shape[0]
        if hasattr(self.regr, "coef_"):
            self.coef_ = self.regr.coef_
        return self

    def predict(self, X):
        # assert hasattr(self.regr, "coef_"), "model must have coef_ attribute"
        return self.regr.predict(X)

    def partial_fit(self, X, y):

        assert hasattr(
            self.regr, "coef_"
        ), "model must be fitted first (i.e have 'coef_' attribute)"
        assert (
            self.n_obs_ is not None
        ), "model must be fitted first (i.e have 'n_obs_' attribute)"

        if len(X.shape) == 1:
            X = X.reshape(1, -1)

        assert X.shape[0] == 1, "X must have one row"

        self.updating_factor_ = self.n_obs_ ** (-self.alpha)

        if isinstance(self.regr, Base):  # nnetsauce model ---

            newX = deepcopy(X)

            if isinstance(
                self.regr, CustomRegressor
            ):  # other nnetsauce model (CustomRegressor) ---
                newX = self.regr.cook_test_set(X=X)
                if isinstance(X, pd.DataFrame):
                    newx = newX.values.ravel()
                else:
                    newx = newX.ravel()

        else:  # an sklearn model ---

            if isinstance(X, pd.DataFrame):
                newx = X.values.ravel()
            else:
                newx = X.ravel()

        new_coef = self.regr.coef_ + self.updating_factor_ * np.dot(
            newx, y - np.dot(newx, self.regr.coef_)
        )
        self.regr.coef_ = _update_mean(self.regr.coef_, self.n_obs_, new_coef)
        self.coef_ = deepcopy(self.regr.coef_)
        self.n_obs_ += 1
        return self



#!/usr/bin/env python

"""
Inductive conformal predictors.
"""

# Authors: Henrik Linusson

from __future__ import division
from collections import defaultdict
from functools import partial

import numpy as np
from sklearn.base import BaseEstimator

from .base import RegressorMixin, ClassifierMixin
from .util import calc_p


# -----------------------------------------------------------------------------
# Base inductive conformal predictor
# -----------------------------------------------------------------------------
class BaseIcp(BaseEstimator):
    """Base class for inductive conformal predictors."""

    def __init__(self, nc_function, condition=None):
        self.cal_x, self.cal_y = None, None
        self.nc_function = nc_function

        # Check if condition-parameter is the default function (i.e.,
        # lambda x: 0). This is so we can safely clone the object without
        # the clone accidentally having self.conditional = True.
        def default_condition(x):
            return 0

        is_default = callable(condition) and (
            condition.__code__.co_code == default_condition.__code__.co_code
        )

        if is_default:
            self.condition = condition
            self.conditional = False
        elif callable(condition):
            self.condition = condition
            self.conditional = True
        else:
            self.condition = lambda x: 0
            self.conditional = False

    def fit(self, x, y):
        """Fit underlying nonconformity scorer.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
                Inputs of examples for fitting the nonconformity scorer.

        y : numpy array of shape [n_samples]
                Outputs of examples for fitting the nonconformity scorer.

        Returns
        -------
        None
        """
        # TODO: incremental?
        self.nc_function.fit(x, y)

    def calibrate(self, x, y, increment=False):
        """Calibrate conformal predictor based on underlying nonconformity
        scorer.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
                Inputs of examples for calibrating the conformal predictor.

        y : numpy array of shape [n_samples, n_features]
                Outputs of examples for calibrating the conformal predictor.

        increment : boolean
                If ``True``, performs an incremental recalibration of the conformal
                predictor. The supplied ``x`` and ``y`` are added to the set of
                previously existing calibration examples, and the conformal
                predictor is then calibrated on both the old and new calibration
                examples.

        Returns
        -------
        None
        """
        self._calibrate_hook(x, y, increment)
        self._update_calibration_set(x, y, increment)

        if self.conditional:
            category_map = np.array(
                [self.condition((x[i, :], y[i])) for i in range(y.size)]
            )
            self.categories = np.unique(category_map)
            self.cal_scores = defaultdict(partial(np.ndarray, 0))

            for cond in self.categories:
                idx = category_map == cond
                cal_scores = self.nc_function.score(self.cal_x[idx, :], self.cal_y[idx])
                self.cal_scores[cond] = np.sort(cal_scores, 0)[::-1]
        else:
            self.categories = np.array([0])
            cal_scores = self.nc_function.score(self.cal_x, self.cal_y)
            self.cal_scores = {0: np.sort(cal_scores, 0)[::-1]}

    def _calibrate_hook(self, x, y, increment):
        pass

    def _update_calibration_set(self, x, y, increment):
        if increment and self.cal_x is not None and self.cal_y is not None:
            self.cal_x = np.vstack([self.cal_x, x])
            self.cal_y = np.hstack([self.cal_y, y])
        else:
            self.cal_x, self.cal_y = x, y


# -----------------------------------------------------------------------------
# Inductive conformal classifier
# -----------------------------------------------------------------------------
class IcpClassifier(BaseIcp, ClassifierMixin):
    """Inductive conformal classifier.

    Parameters
    ----------
    nc_function : BaseScorer
            Nonconformity scorer object used to calculate nonconformity of
            calibration examples and test patterns. Should implement ``fit(x, y)``
            and ``calc_nc(x, y)``.

    smoothing : boolean
            Decides whether to use stochastic smoothing of p-values.

    Attributes
    ----------
    cal_x : numpy array of shape [n_cal_examples, n_features]
            Inputs of calibration set.

    cal_y : numpy array of shape [n_cal_examples]
            Outputs of calibration set.

    nc_function : BaseScorer
            Nonconformity scorer object used to calculate nonconformity scores.

    classes : numpy array of shape [n_classes]
            List of class labels, with indices corresponding to output columns
             of IcpClassifier.predict()

    See also
    --------
    IcpRegressor

    References
    ----------
    .. [1] Papadopoulos, H., & Haralambous, H. (2011). Reliable prediction
            intervals with regression neural networks. Neural Networks, 24(8),
            842-851.

    Examples
    --------
    >>> import numpy as np
    >>> from sklearn.datasets import load_iris
    >>> from sklearn.tree import DecisionTreeClassifier
    >>> from nonconformist.base import ClassifierAdapter
    >>> from nonconformist.icp import IcpClassifier
    >>> from nonconformist.nc import ClassifierNc, MarginErrFunc
    >>> iris = load_iris()
    >>> idx = np.random.permutation(iris.target.size)
    >>> train = idx[:int(idx.size / 3)]
    >>> cal = idx[int(idx.size / 3):int(2 * idx.size / 3)]
    >>> test = idx[int(2 * idx.size / 3):]
    >>> model = ClassifierAdapter(DecisionTreeClassifier())
    >>> nc = ClassifierNc(model, MarginErrFunc())
    >>> icp = IcpClassifier(nc)
    >>> icp.fit(iris.data[train, :], iris.target[train])
    >>> icp.calibrate(iris.data[cal, :], iris.target[cal])
    >>> icp.predict(iris.data[test, :], significance=0.10)
    ...             # doctest: +SKIP
    array([[ True, False, False],
            [False,  True, False],
            ...,
            [False,  True, False],
            [False,  True, False]], dtype=bool)
    """

    def __init__(self, nc_function, condition=None, smoothing=True):
        super(IcpClassifier, self).__init__(nc_function, condition)
        self.classes = None
        self.smoothing = smoothing

    def _calibrate_hook(self, x, y, increment=False):
        self._update_classes(y, increment)

    def _update_classes(self, y, increment):
        if self.classes is None or not increment:
            self.classes = np.unique(y)
        else:
            self.classes = np.unique(np.hstack([self.classes, y]))

    def predict(self, x, significance=None):
        """Predict the output values for a set of input patterns.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
                Inputs of patters for which to predict output values.

        significance : float or None
                Significance level (maximum allowed error rate) of predictions.
                Should be a float between 0 and 1. If ``None``, then the p-values
                are output rather than the predictions.

        Returns
        -------
        p : numpy array of shape [n_samples, n_classes]
                If significance is ``None``, then p contains the p-values for each
                sample-class pair; if significance is a float between 0 and 1, then
                p is a boolean array denoting which labels are included in the
                prediction sets.
        """
        # TODO: if x == self.last_x ...
        n_test_objects = x.shape[0]
        p = np.zeros((n_test_objects, self.classes.size))

        ncal_ngt_neq = self._get_stats(x)

        for i in range(len(self.classes)):
            for j in range(n_test_objects):
                p[j, i] = calc_p(
                    ncal_ngt_neq[j, i, 0],
                    ncal_ngt_neq[j, i, 1],
                    ncal_ngt_neq[j, i, 2],
                    self.smoothing,
                )

        if significance is not None:
            return p > significance
        else:
            return p

    def _get_stats(self, x):
        n_test_objects = x.shape[0]
        ncal_ngt_neq = np.zeros((n_test_objects, self.classes.size, 3))
        for i, c in enumerate(self.classes):
            test_class = np.zeros(x.shape[0], dtype=self.classes.dtype)
            test_class.fill(c)

            # TODO: maybe calculate p-values using cython or similar
            # TODO: interpolated p-values

            # TODO: nc_function.calc_nc should take X * {y1, y2, ... ,yn}
            test_nc_scores = self.nc_function.score(x, test_class)
            for j, nc in enumerate(test_nc_scores):
                cal_scores = self.cal_scores[self.condition((x[j, :], c))][::-1]
                n_cal = cal_scores.size

                idx_left = np.searchsorted(cal_scores, nc, "left")
                idx_right = np.searchsorted(cal_scores, nc, "right")

                ncal_ngt_neq[j, i, 0] = n_cal
                ncal_ngt_neq[j, i, 1] = n_cal - idx_right
                ncal_ngt_neq[j, i, 2] = idx_right - idx_left

        return ncal_ngt_neq

    def predict_conf(self, x):
        """Predict the output values for a set of input patterns, using
        the confidence-and-credibility output scheme.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
                Inputs of patters for which to predict output values.

        Returns
        -------
        p : numpy array of shape [n_samples, 3]
                p contains three columns: the first column contains the most
                likely class for each test pattern; the second column contains
                the confidence in the predicted class label, and the third column
                contains the credibility of the prediction.
        """
        p = self.predict(x, significance=None)
        label = p.argmax(axis=1)
        credibility = p.max(axis=1)
        for i, idx in enumerate(label):
            p[i, idx] = -np.inf
        confidence = 1 - p.max(axis=1)

        return np.array([label, confidence, credibility]).T


# -----------------------------------------------------------------------------
# Inductive conformal regressor
# -----------------------------------------------------------------------------
class IcpRegressor(BaseIcp, RegressorMixin):
    """Inductive conformal regressor.

    Parameters
    ----------
    nc_function : BaseScorer
            Nonconformity scorer object used to calculate nonconformity of
            calibration examples and test patterns. Should implement ``fit(x, y)``,
            ``calc_nc(x, y)`` and ``predict(x, nc_scores, significance)``.

    Attributes
    ----------
    cal_x : numpy array of shape [n_cal_examples, n_features]
            Inputs of calibration set.

    cal_y : numpy array of shape [n_cal_examples]
            Outputs of calibration set.

    nc_function : BaseScorer
            Nonconformity scorer object used to calculate nonconformity scores.

    See also
    --------
    IcpClassifier

    References
    ----------
    .. [1] Papadopoulos, H., Proedrou, K., Vovk, V., & Gammerman, A. (2002).
            Inductive confidence machines for regression. In Machine Learning: ECML
            2002 (pp. 345-356). Springer Berlin Heidelberg.

    .. [2] Papadopoulos, H., & Haralambous, H. (2011). Reliable prediction
            intervals with regression neural networks. Neural Networks, 24(8),
            842-851.

    Examples
    --------
    >>> import numpy as np
    >>> from sklearn.datasets import load_boston
    >>> from sklearn.tree import DecisionTreeRegressor
    >>> from nonconformist.base import RegressorAdapter
    >>> from nonconformist.icp import IcpRegressor
    >>> from nonconformist.nc import RegressorNc, AbsErrorErrFunc
    >>> boston = load_boston()
    >>> idx = np.random.permutation(boston.target.size)
    >>> train = idx[:int(idx.size / 3)]
    >>> cal = idx[int(idx.size / 3):int(2 * idx.size / 3)]
    >>> test = idx[int(2 * idx.size / 3):]
    >>> model = RegressorAdapter(DecisionTreeRegressor())
    >>> nc = RegressorNc(model, AbsErrorErrFunc())
    >>> icp = IcpRegressor(nc)
    >>> icp.fit(boston.data[train, :], boston.target[train])
    >>> icp.calibrate(boston.data[cal, :], boston.target[cal])
    >>> icp.predict(boston.data[test, :], significance=0.10)
    ...     # doctest: +SKIP
    array([[  5. ,  20.6],
            [ 15.5,  31.1],
            ...,
            [ 14.2,  29.8],
            [ 11.6,  27.2]])
    """

    def __init__(self, nc_function, condition=None):
        super(IcpRegressor, self).__init__(nc_function, condition)

    def predict(self, x, significance=None):
        """Predict the output values for a set of input patterns.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
                Inputs of patters for which to predict output values.

        significance : float
                Significance level (maximum allowed error rate) of predictions.
                Should be a float between 0 and 1. If ``None``, then intervals for
                all significance levels (0.01, 0.02, ..., 0.99) are output in a
                3d-matrix.

        Returns
        -------
        p : numpy array of shape [n_samples, 2] or [n_samples, 2, 99}
                If significance is ``None``, then p contains the interval (minimum
                and maximum boundaries) for each test pattern, and each significance
                level (0.01, 0.02, ..., 0.99). If significance is a float between
                0 and 1, then p contains the prediction intervals (minimum and
                maximum	boundaries) for the set of test patterns at the chosen
                significance level.
        """
        # TODO: interpolated p-values

        n_significance = 99 if significance is None else np.array(significance).size

        if n_significance > 1:
            prediction = np.zeros((x.shape[0], 2, n_significance))
        else:
            prediction = np.zeros((x.shape[0], 2))

        condition_map = np.array(
            [self.condition((x[i, :], None)) for i in range(x.shape[0])]
        )

        for condition in self.categories:
            idx = condition_map == condition
            if np.sum(idx) > 0:
                p = self.nc_function.predict(
                    x[idx, :], self.cal_scores[condition], significance
                )
                if n_significance > 1:
                    prediction[idx, :, :] = p
                else:
                    prediction[idx, :] = p

        return prediction


class OobCpClassifier(IcpClassifier):
    def __init__(self, nc_function, condition=None, smoothing=True):
        super(OobCpClassifier, self).__init__(nc_function, condition, smoothing)

    def fit(self, x, y):
        super(OobCpClassifier, self).fit(x, y)
        super(OobCpClassifier, self).calibrate(x, y, False)

    def calibrate(self, x, y, increment=False):
        # Should throw exception (or really not be implemented for oob)
        pass


class OobCpRegressor(IcpRegressor):
    def __init__(self, nc_function, condition=None):
        super(OobCpRegressor, self).__init__(nc_function, condition)

    def fit(self, x, y):
        super(OobCpRegressor, self).fit(x, y)
        super(OobCpRegressor, self).calibrate(x, y, False)

    def calibrate(self, x, y, increment=False):
        # Should throw exception (or really not be implemented for oob)
        pass



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import gaussian_kde, norm
from scipy.interpolate import interp1d


def simulate_distribution(data, method="bootstrap", num_samples=1000, **kwargs):
    """
    Simulate the distribution of an input vector using various methods.

    Parameters:
        data (array-like): Input vector of data.
        method (str): Method for simulation:
                      - 'bootstrap': Bootstrap resampling.
                      - 'kde': Kernel Density Estimation.
                      - 'normal': Normal distribution.
                      - 'ecdf': Empirical CDF-based sampling.
                      - 'permutation': Permutation resampling.
                      - 'smooth-bootstrap': Smoothed bootstrap with added noise.
        num_samples (int): Number of samples to generate.
        kwargs: Additional parameters for specific methods:
                - kde_bandwidth (str or float): Bandwidth for KDE ('scott', 'silverman', or float).
                - dist (str): Parametric distribution type ('normal').
                - noise_std (float): Noise standard deviation for smoothed bootstrap.

    Returns:
        np.ndarray: Simulated distribution samples.
    """
    assert method in [
        "bootstrap",
        "kde",
        "parametric",
        "ecdf",
        "permutation",
        "smooth-bootstrap",
    ], f"Unknown method '{method}'. Choose from 'bootstrap', 'kde', 'parametric', 'ecdf', 'permutation', or 'smooth_bootstrap'."

    data = np.array(data)

    if method == "bootstrap":
        simulated_data = np.random.choice(data, size=num_samples, replace=True)

    elif method == "kde":
        kde_bandwidth = kwargs.get("kde_bandwidth", "scott")
        kde = gaussian_kde(data, bw_method=kde_bandwidth)
        simulated_data = kde.resample(num_samples).flatten()

    elif method == "normal":
        mean, std = np.mean(data), np.std(data)
        simulated_data = np.random.normal(mean, std, size=num_samples)

    elif method == "ecdf":
        data = np.sort(data)
        ecdf_y = np.arange(1, len(data) + 1) / len(data)
        inverse_cdf = interp1d(
            ecdf_y, data, bounds_error=False, fill_value=(data[0], data[-1]))
        random_uniform = np.random.uniform(0, 1, size=num_samples)
        simulated_data = inverse_cdf(random_uniform)

    elif method == "permutation":
        simulated_data = np.random.permutation(data)
        while len(simulated_data) < num_samples:
            simulated_data = np.concatenate(
                [simulated_data, np.random.permutation(data)])
        simulated_data = simulated_data[:num_samples]

    elif method == "smooth_bootstrap":
        noise_std = kwargs.get("noise_std", 0.1)
        bootstrap_samples = np.random.choice(
            data, size=num_samples, replace=True)
        noise = np.random.normal(0, noise_std, size=num_samples)
        simulated_data = bootstrap_samples + noise

    else:
        raise ValueError(
            f"Unknown method '{method}'. Choose from 'bootstrap', 'kde', 'parametric', 'ecdf', 'permutation', or 'smooth_bootstrap'."
        )

    return simulated_data


def simulate_replications(data, method="kde", num_replications=10, n_obs=None, **kwargs):
    """
    Create multiple replications of the input's distribution using a specified simulation method.

    Parameters:
        data (array-like): Input vector of data.
        method (str): Method for simulation:
                      - 'bootstrap': Bootstrap resampling.
                      - 'kde': Kernel Density Estimation.
                      - 'normal': Parametric distribution fitting.
                      - 'ecdf': Empirical CDF-based sampling.
                      - 'permutation': Permutation resampling.
                      - 'smooth_bootstrap': Smoothed bootstrap with added noise.
        num_samples (int): Number of samples in each replication.
        num_replications (int): Number of replications to generate.
        n_obs (int): Number of observations to generate for each replication.
        kwargs: Additional parameters for specific methods.

    Returns:
        pd.DataFrame: A DataFrame where each column represents a replication.
    """

    num_samples = len(data)

    replications = []

    for _ in range(num_replications):
        simulated_data = simulate_distribution(
            data, method=method, num_samples=num_samples, **kwargs)
        replications.append(simulated_data)

    # Combine replications into a DataFrame
    replications_df = pd.DataFrame(replications).transpose()
    replications_df.columns = [
        f"Replication_{i+1}" for i in range(num_replications)]

    # If n_obs is specified, sample n_obs from each replication
    if n_obs is not None:
        replications_df = replications_df.sample(
            n=n_obs, replace=True, random_state=42).reset_index(drop=True)
        return replications_df.values

    return replications_df.values



from .base.base import Base
from .base.baseRegressor import BaseRegressor
from .boosting.adaBoostClassifier import AdaBoostClassifier
from .custom.customClassifier import CustomClassifier
from .custom.customRegressor import CustomRegressor
from .datasets import Downloader
from .deep.deepClassifier import DeepClassifier
from .deep.deepRegressor import DeepRegressor
from .deep.deepMTS import DeepMTS
from .glm.glmClassifier import GLMClassifier
from .glm.glmRegressor import GLMRegressor
from .kernel.kernel import KernelRidge
from .lazypredict.lazydeepClassifier import LazyDeepClassifier, LazyClassifier
from .lazypredict.lazydeepRegressor import LazyDeepRegressor, LazyRegressor
from .lazypredict.lazydeepClassifier import LazyDeepClassifier
from .lazypredict.lazydeepRegressor import LazyDeepRegressor
from .lazypredict.lazydeepMTS import LazyDeepMTS, LazyMTS
from .mts.mts import MTS
from .mts.mlarch import MLARCH
from .mts.classical import ClassicalMTS
from .multitask.multitaskClassifier import MultitaskClassifier
from .multitask.simplemultitaskClassifier import SimpleMultitaskClassifier
from .neuralnet.neuralnetregression import NeuralNetRegressor
from .neuralnet.neuralnetclassification import NeuralNetClassifier
from .optimizers.optimizer import Optimizer
from .predictioninterval import PredictionInterval
from .quantile.quantileregression import QuantileRegressor
from .quantile.quantileclassification import QuantileClassifier
from .randombag.randomBagClassifier import RandomBagClassifier
from .randombag.randomBagRegressor import RandomBagRegressor
from .ridge.ridge import RidgeRegressor
from .ridge2.ridge2Classifier import Ridge2Classifier
from .ridge2.ridge2Regressor import Ridge2Regressor
from .ridge2.ridge2MultitaskClassifier import Ridge2MultitaskClassifier
from .rvfl.bayesianrvflRegressor import BayesianRVFLRegressor
from .rvfl.bayesianrvfl2Regressor import BayesianRVFL2Regressor
from .sampling import SubSampler
from .updater import RegressorUpdater, ClassifierUpdater
from .votingregressor import MedianVotingRegressor

__all__ = [
    "AdaBoostClassifier",
    "Base",
    "BaseRegressor",
    "BayesianRVFLRegressor",
    "BayesianRVFL2Regressor",
    "ClassicalMTS",
    "CustomClassifier",
    "CustomRegressor",
    "DeepClassifier",
    "DeepRegressor",
    "DeepMTS",
    "Downloader",
    "GLMClassifier",
    "GLMRegressor",
    "KernelRidge",
    "LazyClassifier",
    "LazyRegressor",
    "LazyDeepClassifier",
    "LazyDeepRegressor",
    "LazyMTS",
    "LazyDeepMTS",
    "MLARCH",
    "MedianVotingRegressor",
    "MTS",
    "MultitaskClassifier",
    "NeuralNetRegressor",
    "NeuralNetClassifier",
    "PredictionInterval",
    "SimpleMultitaskClassifier",
    "Optimizer",
    "QuantileRegressor",
    "QuantileClassifier",
    "RandomBagRegressor",
    "RandomBagClassifier",
    "RegressorUpdater",
    "ClassifierUpdater",
    "RidgeRegressor",
    "Ridge2Regressor",
    "Ridge2Classifier",
    "Ridge2MultitaskClassifier",
    "SubSampler",
]



import numpy as np
import pandas as pd
import sklearn.metrics as skm2
from copy import deepcopy
from tqdm import tqdm
from sklearn.base import RegressorMixin
from ..custom import CustomRegressor
from ..predictioninterval import PredictionInterval
from ..utils import matrixops as mo


class DeepRegressor(CustomRegressor, RegressorMixin):
    """
    Deep Regressor

    Parameters:

        obj: an object
            A base learner, see also https://www.researchgate.net/publication/380701207_Deep_Quasi-Randomized_neural_Networks_for_classification

        verbose : int, optional (default=0)
            Monitor progress when fitting.

        n_layers: int (default=2)
            Number of layers. `n_layers = 1` is a simple `CustomRegressor`

        All the other parameters are nnetsauce `CustomRegressor`'s

    Examples:

        ```python
        import nnetsauce as ns
        from sklearn.datasets import load_diabetes
        from sklearn.model_selection import train_test_split
        from sklearn.linear_model import RidgeCV
        data = load_diabetes()
        X = data.data
        y= data.target
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=.2, random_state=123)
        obj = RidgeCV()
        clf = ns.DeepRegressor(obj)
        clf.fit(X_train, y_train)
        print(clf.score(clf.predict(X_test), y_test))
        ```

    """

    def __init__(
        self,
        obj,
        # Defining depth
        n_layers=2,
        verbose=0,
        # CustomRegressor attributes
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        level=None,
        pi_method="splitconformal",
        seed=123,
        backend="cpu",
    ):
        super().__init__(
            obj=obj,
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            level=level,
            pi_method=pi_method,
            seed=seed,
            backend=backend,
        )

        assert n_layers >= 1, "must have n_layers >= 1"

        self.stacked_obj = deepcopy(obj)
        self.verbose = verbose
        self.n_layers = n_layers
        self.level = level
        self.pi_method = pi_method
        self.coef_ = None

    def fit(self, X, y, **kwargs):
        """Fit Regression algorithms to X and y.
        Parameters
        ----------
        X : array-like,
            Training vectors, where rows is the number of samples
            and columns is the number of features.
        y : array-like,
            Training vectors, where rows is the number of samples
            and columns is the number of features.
        **kwargs: dict
            Additional parameters to be passed to the fit method
            of the base learner. For example, `sample_weight`.
        Returns
        -------
        A fitted object
        """

        if isinstance(X, np.ndarray):
            X = pd.DataFrame(X)

        # init layer
        self.stacked_obj = CustomRegressor(
            obj=self.stacked_obj,
            n_hidden_features=self.n_hidden_features,
            activation_name=self.activation_name,
            a=self.a,
            nodes_sim=self.nodes_sim,
            bias=self.bias,
            dropout=self.dropout,
            direct_link=self.direct_link,
            n_clusters=self.n_clusters,
            cluster_encode=self.cluster_encode,
            type_clust=self.type_clust,
            type_scaling=self.type_scaling,
            col_sample=self.col_sample,
            row_sample=self.row_sample,
            seed=self.seed,
            backend=self.backend,
        )

        if self.verbose > 0:
            iterator = tqdm(range(self.n_layers - 1))
        else:
            iterator = range(self.n_layers - 1)

        for _ in iterator:
            self.stacked_obj = deepcopy(
                CustomRegressor(
                    obj=self.stacked_obj,
                    n_hidden_features=self.n_hidden_features,
                    activation_name=self.activation_name,
                    a=self.a,
                    nodes_sim=self.nodes_sim,
                    bias=self.bias,
                    dropout=self.dropout,
                    direct_link=self.direct_link,
                    n_clusters=self.n_clusters,
                    cluster_encode=self.cluster_encode,
                    type_clust=self.type_clust,
                    type_scaling=self.type_scaling,
                    col_sample=self.col_sample,
                    row_sample=self.row_sample,
                    seed=self.seed,
                    backend=self.backend,
                )
            )

        self.stacked_obj.fit(X, y, **kwargs)

        if self.level is not None:
            self.stacked_obj = PredictionInterval(
                obj=self.stacked_obj, method=self.pi_method, level=self.level
            )

        if hasattr(self.stacked_obj, "clustering_obj_"):
            self.clustering_obj_ = self.stacked_obj.clustering_obj_

        if hasattr(self.stacked_obj, "coef_"):
            self.coef_ = self.stacked_obj.coef_

        if hasattr(self.stacked_obj, "scaler_"):
            self.scaler_ = self.stacked_obj.scaler_

        if hasattr(self.stacked_obj, "nn_scaler_"):
            self.nn_scaler_ = self.stacked_obj.nn_scaler_

        if hasattr(self.stacked_obj, "clustering_scaler_"):
            self.clustering_scaler_ = self.stacked_obj.clustering_scaler_

        return self

    def partial_fit(self, X, y, **kwargs):
        """Fit Regression algorithms to X and y.
        Parameters
        ----------
        X : array-like,
            Training vectors, where rows is the number of samples
            and columns is the number of features.
        y : array-like,
            Training vectors, where rows is the number of samples
            and columns is the number of features.
        **kwargs: dict
            Additional parameters to be passed to the fit method
            of the base learner. For example, `sample_weight`.
        Returns
        -------
        A fitted object
        """
        assert hasattr(self, "stacked_obj"), "model must be fitted first"
        current_obj = self.stacked_obj
        for _ in range(self.n_layers):
            try:
                input_X = current_obj.obj.cook_test_set(X)
                current_obj.obj.partial_fit(input_X, y, **kwargs)
                try:
                    current_obj = current_obj.obj
                except AttributeError:
                    pass
            except ValueError as e:
                print(e)
                pass
        return self

    def predict(self, X, **kwargs):
        if self.level is not None:
            return self.stacked_obj.predict(X, return_pi=True)
        return self.stacked_obj.predict(X, **kwargs)

    def score(self, X, y, scoring=None):
        return self.stacked_obj.score(X, y, scoring)



#!/usr/bin/env python

"""
Aggregated conformal predictors
"""

# Authors: Henrik Linusson

import numpy as np
from sklearn.model_selection import KFold, StratifiedKFold
from sklearn.model_selection import ShuffleSplit, StratifiedShuffleSplit
from sklearn.base import clone
from nonconformist.base import BaseEstimator
from nonconformist.util import calc_p


# -----------------------------------------------------------------------------
# Sampling strategies
# -----------------------------------------------------------------------------
class BootstrapSampler(object):
    """Bootstrap sampler.

    See also
    --------
    CrossSampler, RandomSubSampler

    Examples
    --------
    """

    def gen_samples(self, y, n_samples, problem_type):
        for i in range(n_samples):
            idx = np.array(range(y.size))
            train = np.random.choice(y.size, y.size, replace=True)
            cal_mask = np.array(np.ones(idx.size), dtype=bool)
            for j in train:
                cal_mask[j] = False
            cal = idx[cal_mask]

            yield train, cal


class CrossSampler(object):
    """Cross-fold sampler.

    See also
    --------
    BootstrapSampler, RandomSubSampler

    Examples
    --------
    """

    def gen_samples(self, y, n_samples, problem_type):
        if problem_type == "classification":
            folds = StratifiedKFold(y, n_folds=n_samples)
        else:
            folds = KFold(y.size, n_folds=n_samples)
        for train, cal in folds:
            yield train, cal


class RandomSubSampler(object):
    """Random subsample sampler.

    Parameters
    ----------
    calibration_portion : float
            Ratio (0-1) of examples to use for calibration.

    See also
    --------
    BootstrapSampler, CrossSampler

    Examples
    --------
    """

    def __init__(self, calibration_portion=0.3):
        self.cal_portion = calibration_portion

    def gen_samples(self, y, n_samples, problem_type):
        if problem_type == "classification":
            splits = StratifiedShuffleSplit(
                y, n_iter=n_samples, test_size=self.cal_portion
            )
        else:
            splits = ShuffleSplit(y.size, n_iter=n_samples, test_size=self.cal_portion)

        for train, cal in splits:
            yield train, cal


# -----------------------------------------------------------------------------
# Conformal ensemble
# -----------------------------------------------------------------------------
class AggregatedCp(BaseEstimator):
    """Aggregated conformal predictor.

    Combines multiple IcpClassifier or IcpRegressor predictors into an
    aggregated model.

    Parameters
    ----------
    predictor : object
            Prototype conformal predictor (e.g. IcpClassifier or IcpRegressor)
            used for defining conformal predictors included in the aggregate model.

    sampler : object
            Sampler object used to generate training and calibration examples
            for the underlying conformal predictors.

    aggregation_func : callable
            Function used to aggregate the predictions of the underlying
            conformal predictors. Defaults to ``numpy.mean``.

    n_models : int
            Number of models to aggregate.

    Attributes
    ----------
    predictor : object
            Prototype conformal predictor.

    predictors : list
            List of underlying conformal predictors.

    sampler : object
            Sampler object used to generate training and calibration examples.

    agg_func : callable
            Function used to aggregate the predictions of the underlying
            conformal predictors

    References
    ----------
    .. [1] Vovk, V. (2013). Cross-conformal predictors. Annals of Mathematics
            and Artificial Intelligence, 1-20.

    .. [2] Carlsson, L., Eklund, M., & Norinder, U. (2014). Aggregated
            Conformal Prediction. In Artificial Intelligence Applications and
            Innovations (pp. 231-240). Springer Berlin Heidelberg.

    Examples
    --------
    """

    def __init__(
        self,
        predictor,
        sampler=BootstrapSampler(),
        aggregation_func=None,
        n_models=10,
    ):
        self.predictors = []
        self.n_models = n_models
        self.predictor = predictor
        self.sampler = sampler

        if aggregation_func is not None:
            self.agg_func = aggregation_func
        else:
            self.agg_func = lambda x: np.mean(x, axis=2)

    def fit(self, x, y):
        """Fit underlying conformal predictors.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
                Inputs of examples for fitting the underlying conformal predictors.

        y : numpy array of shape [n_samples]
                Outputs of examples for fitting the underlying conformal predictors.

        Returns
        -------
        None
        """
        self.n_train = y.size
        self.predictors = []
        idx = np.random.permutation(y.size)
        x, y = x[idx, :], y[idx]
        problem_type = self.predictor.__class__.get_problem_type()
        samples = self.sampler.gen_samples(y, self.n_models, problem_type)
        for train, cal in samples:
            predictor = clone(self.predictor)
            predictor.fit(x[train, :], y[train])
            predictor.calibrate(x[cal, :], y[cal])
            self.predictors.append(predictor)

        if problem_type == "classification":
            self.classes = self.predictors[0].classes

    def predict(self, x, significance=None):
        """Predict the output values for a set of input patterns.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
                Inputs of patters for which to predict output values.

        significance : float or None
                Significance level (maximum allowed error rate) of predictions.
                Should be a float between 0 and 1. If ``None``, then the p-values
                are output rather than the predictions. Note: ``significance=None``
                is applicable to classification problems only.

        Returns
        -------
        p : numpy array of shape [n_samples, n_classes] or [n_samples, 2]
                For classification problems: If significance is ``None``, then p
                contains the p-values for each sample-class pair; if significance
                is a float between 0 and 1, then p is a boolean array denoting
                which labels are included in the prediction sets.

                For regression problems: Prediction interval (minimum and maximum
                boundaries) for the set of test patterns.
        """
        is_regression = self.predictor.__class__.get_problem_type() == "regression"

        n_examples = x.shape[0]

        if is_regression and significance is None:
            signs = np.arange(0.01, 1.0, 0.01)
            pred = np.zeros((n_examples, 2, signs.size))
            for i, s in enumerate(signs):
                predictions = np.dstack([p.predict(x, s) for p in self.predictors])
                predictions = self.agg_func(predictions)
                pred[:, :, i] = predictions
            return pred
        else:

            def f(p, x):
                return p.predict(x, significance if is_regression else None)

            predictions = np.dstack([f(p, x) for p in self.predictors])
            predictions = self.agg_func(predictions)

            if significance and not is_regression:
                return predictions >= significance
            else:
                return predictions


class CrossConformalClassifier(AggregatedCp):
    """Cross-conformal classifier.

    Combines multiple IcpClassifiers into a cross-conformal classifier.

    Parameters
    ----------
    predictor : object
            Prototype conformal predictor (e.g. IcpClassifier or IcpRegressor)
            used for defining conformal predictors included in the aggregate model.

    aggregation_func : callable
            Function used to aggregate the predictions of the underlying
            conformal predictors. Defaults to ``numpy.mean``.

    n_models : int
            Number of models to aggregate.

    Attributes
    ----------
    predictor : object
            Prototype conformal predictor.

    predictors : list
            List of underlying conformal predictors.

    sampler : object
            Sampler object used to generate training and calibration examples.

    agg_func : callable
            Function used to aggregate the predictions of the underlying
            conformal predictors

    References
    ----------
    .. [1] Vovk, V. (2013). Cross-conformal predictors. Annals of Mathematics
            and Artificial Intelligence, 1-20.

    Examples
    --------
    """

    def __init__(self, predictor, n_models=10):
        super(CrossConformalClassifier, self).__init__(
            predictor, CrossSampler(), n_models
        )

    def predict(self, x, significance=None):
        ncal_ngt_neq = np.stack([p._get_stats(x) for p in self.predictors], axis=3)
        ncal_ngt_neq = ncal_ngt_neq.sum(axis=3)

        p = calc_p(
            ncal_ngt_neq[:, :, 0],
            ncal_ngt_neq[:, :, 1],
            ncal_ngt_neq[:, :, 2],
            smoothing=self.predictors[0].smoothing,
        )

        if significance:
            return p > significance
        else:
            return p


class BootstrapConformalClassifier(AggregatedCp):
    """Bootstrap conformal classifier.

    Combines multiple IcpClassifiers into a bootstrap conformal classifier.

    Parameters
    ----------
    predictor : object
            Prototype conformal predictor (e.g. IcpClassifier or IcpRegressor)
            used for defining conformal predictors included in the aggregate model.

    aggregation_func : callable
            Function used to aggregate the predictions of the underlying
            conformal predictors. Defaults to ``numpy.mean``.

    n_models : int
            Number of models to aggregate.

    Attributes
    ----------
    predictor : object
            Prototype conformal predictor.

    predictors : list
            List of underlying conformal predictors.

    sampler : object
            Sampler object used to generate training and calibration examples.

    agg_func : callable
            Function used to aggregate the predictions of the underlying
            conformal predictors

    References
    ----------
    .. [1] Vovk, V. (2013). Cross-conformal predictors. Annals of Mathematics
            and Artificial Intelligence, 1-20.

    Examples
    --------
    """

    def __init__(self, predictor, n_models=10):
        super(BootstrapConformalClassifier, self).__init__(
            predictor, BootstrapSampler(), n_models
        )

    def predict(self, x, significance=None):
        ncal_ngt_neq = np.stack([p._get_stats(x) for p in self.predictors], axis=3)
        ncal_ngt_neq = ncal_ngt_neq.sum(axis=3)

        p = calc_p(
            ncal_ngt_neq[:, :, 0] + ncal_ngt_neq[:, :, 0] / self.n_train,
            ncal_ngt_neq[:, :, 1] + ncal_ngt_neq[:, :, 0] / self.n_train,
            ncal_ngt_neq[:, :, 2],
            smoothing=self.predictors[0].smoothing,
        )

        if significance:
            return p > significance
        else:
            return p



#!/usr/bin/env python

"""
docstring
"""

# Authors: Henrik Linusson

import abc
import numpy as np

from sklearn.base import BaseEstimator


class RegressorMixin(object):
    def __init__(self):
        super(RegressorMixin, self).__init__()

    @classmethod
    def get_problem_type(cls):
        return "regression"


class ClassifierMixin(object):
    def __init__(self):
        super(ClassifierMixin, self).__init__()

    @classmethod
    def get_problem_type(cls):
        return "classification"


class BaseModelAdapter(BaseEstimator):
    __metaclass__ = abc.ABCMeta

    def __init__(self, model, fit_params=None):
        super(BaseModelAdapter, self).__init__()

        self.model = model
        self.last_x, self.last_y = None, None
        self.clean = False
        self.fit_params = {} if fit_params is None else fit_params

    def fit(self, x, y):
        """Fits the model.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
                Inputs of examples for fitting the model.

        y : numpy array of shape [n_samples]
                Outputs of examples for fitting the model.

        Returns
        -------
        None
        """

        self.model.fit(x, y, **self.fit_params)
        self.clean = False

    def predict(self, x):
        """Returns the prediction made by the underlying model.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
                Inputs of test examples.

        Returns
        -------
        y : numpy array of shape [n_samples]
                Predicted outputs of test examples.
        """
        if (
            not self.clean
            or self.last_x is None
            or self.last_y is None
            or not np.array_equal(self.last_x, x)
        ):
            self.last_x = x
            self.last_y = self._underlying_predict(x)
            self.clean = True

        return self.last_y.copy()

    @abc.abstractmethod
    def _underlying_predict(self, x):
        """Produces a prediction using the encapsulated model.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
                Inputs of test examples.

        Returns
        -------
        y : numpy array of shape [n_samples]
                Predicted outputs of test examples.
        """
        pass


class ClassifierAdapter(BaseModelAdapter):
    def __init__(self, model, fit_params=None):
        super(ClassifierAdapter, self).__init__(model, fit_params)

    def _underlying_predict(self, x):
        return self.model.predict_proba(x)


class RegressorAdapter(BaseModelAdapter):
    def __init__(self, model, fit_params=None):
        super(RegressorAdapter, self).__init__(model, fit_params)

    def _underlying_predict(self, x):
        return self.model.predict(x)


class OobMixin(object):
    def __init__(self, model, fit_params=None):
        super(OobMixin, self).__init__(model, fit_params)
        self.train_x = None

    def fit(self, x, y):
        super(OobMixin, self).fit(x, y)
        self.train_x = x

    def _underlying_predict(self, x):
        # TODO: sub-sampling of ensemble for test patterns
        oob = x == self.train_x

        if hasattr(oob, "all"):
            oob = oob.all()

        if oob:
            return self._oob_prediction()
        else:
            return super(OobMixin, self)._underlying_predict(x)


class OobClassifierAdapter(OobMixin, ClassifierAdapter):
    def __init__(self, model, fit_params=None):
        super(OobClassifierAdapter, self).__init__(model, fit_params)

    def _oob_prediction(self):
        return self.model.oob_decision_function_


class OobRegressorAdapter(OobMixin, RegressorAdapter):
    def __init__(self, model, fit_params=None):
        super(OobRegressorAdapter, self).__init__(model, fit_params)

    def _oob_prediction(self):
        return self.model.oob_prediction_



# Authors: Thierry Moudiki
#
# License: BSD 3 Clear

import numpy as np
from joblib import Parallel, delayed
from tqdm import tqdm
from ..utils.misc import flatten, is_factor


import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.mixture import GaussianMixture
from sklearn.cluster import DBSCAN
from scipy.spatial.distance import cdist
from scipy.optimize import linear_sum_assignment


import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.mixture import GaussianMixture
from sklearn.cluster import DBSCAN
from scipy.spatial.distance import cdist
from scipy.optimize import linear_sum_assignment


def subsample_indices(
    data,
    method="random",
    fraction=0.1,
    label_col=None,
    n_components=10,
    eps=0.5,
    min_samples=5,
    random_state=42,
):
    """
    Subsample row indices from a dataset while preserving statistical properties.

    Parameters:
        data (pd.DataFrame or np.ndarray): The input data (Pandas DataFrame or NumPy array).
        method (str): Sampling method. Options: 'random', 'stratified', 'gmm', 'dbscan', 'optimal_transport'.
        fraction (float): Fraction of rows to sample (e.g., 0.1 for 10%).
        label_col (str or int): Column name/index for stratified sampling (required if method='stratified').
                                For NumPy arrays, provide the column index.
        n_components (int): Number of components for GMM-based sampling (used if method='gmm').
        eps (float): The maximum distance between two samples for them to be considered in the same neighborhood (used if method='dbscan').
        min_samples (int): The number of samples in a neighborhood for a point to be considered a core point (used if method='dbscan').
        random_state (int): Random seed for reproducibility.

    Returns:
        list: Indices of the subsampled rows.
    """
    # Check if the input is a Pandas DataFrame or NumPy array
    if isinstance(data, pd.DataFrame):
        is_dataframe = True
        if label_col is not None and label_col not in data.columns:
            raise ValueError(f"Column '{label_col}' does not exist in the DataFrame.")
        X = data.drop(columns=[label_col] if label_col else []).values
        labels = data[label_col].values if label_col else None
    elif isinstance(data, np.ndarray):
        is_dataframe = False
        if label_col is not None:
            if (
                not isinstance(label_col, int)
                or label_col < 0
                or label_col >= data.shape[1]
            ):
                raise ValueError(
                    "For NumPy arrays, 'label_col' must be a valid column index."
                )
            labels = data[:, label_col]
            X = np.delete(data, label_col, axis=1)
        else:
            X = data
            labels = None
    else:
        raise TypeError("Input data must be a Pandas DataFrame or a NumPy array.")

    if method == "random":
        # Simple random sampling
        rng = np.random.default_rng(random_state)
        subsample_indices = rng.choice(
            data.shape[0], size=int(data.shape[0] * fraction), replace=False
        )
        return subsample_indices.tolist()

    elif method == "stratified":
        if labels is None:
            raise ValueError("For stratified sampling, 'label_col' must be specified.")
        # Stratified sampling
        _, _, _, subsample_indices = train_test_split(
            np.arange(data.shape[0]),
            labels,
            test_size=fraction,
            stratify=labels,
            random_state=random_state,
        )
        return subsample_indices.tolist()

    elif method == "gmm":
        # GMM-based sampling
        gmm = GaussianMixture(n_components=n_components, random_state=random_state).fit(
            X
        )

        # Assign each row to a cluster
        clusters = gmm.predict(X)

        # Sample proportionally from each cluster
        rng = np.random.default_rng(random_state)
        subsample_indices = []
        for cluster in np.unique(clusters):
            cluster_indices = np.where(clusters == cluster)[0]
            subsample_size = max(1, int(len(cluster_indices) * fraction))
            subsample_indices.extend(
                rng.choice(cluster_indices, size=subsample_size, replace=False)
            )
        return subsample_indices

    elif method == "dbscan":
        # DBSCAN-based sampling
        dbscan = DBSCAN(eps=eps, min_samples=min_samples).fit(X)

        # Assign each row to a cluster (or -1 for noise)
        clusters = dbscan.labels_

        # Exclude noise points (-1) and sample proportionally from each cluster
        valid_clusters = np.where(clusters != -1)[0]
        if len(valid_clusters) == 0:
            raise ValueError(
                "DBSCAN identified only noise points. Adjust `eps` or `min_samples`."
            )

        rng = np.random.default_rng(random_state)
        subsample_indices = []
        for cluster in np.unique(clusters[valid_clusters]):
            cluster_indices = np.where(clusters == cluster)[0]
            subsample_size = max(1, int(len(cluster_indices) * fraction))
            subsample_indices.extend(
                rng.choice(cluster_indices, size=subsample_size, replace=False)
            )
        return subsample_indices

    elif method == "optimal_transport":
        # Optimal transport-based sampling
        distances = cdist(X, X, metric="euclidean")

        # Solve the optimal transport problem
        row_ind, col_ind = linear_sum_assignment(distances)

        # Select the subsample based on the solution
        subsample_size = int(data.shape[0] * fraction)
        subsample_indices = col_ind[:subsample_size]
        return subsample_indices.tolist()

    else:
        raise ValueError(
            "Invalid sampling method. Choose from 'random', 'stratified', 'gmm', 'dbscan', or 'optimal_transport'."
        )


# Example usage:
# df = pd.DataFrame({'feature1': np.random.randn(1000), 'feature2': np.random.randn(1000), 'label': np.random.choice([0, 1], size=1000)})
# indices = subsample_indices(df, method='dbscan', fraction=0.1, eps=0.5, min_samples=5)
# print(indices)


def dosubsample(y, row_sample=0.8, seed=123, n_jobs=None, verbose=False):
    index = []
    assert (row_sample < 1) & (row_sample >= 0), "'row_sample' must be < 1 and >= 0"
    n_obs = len(y)
    n_obs_out = np.ceil(n_obs * row_sample)

    # preproc -----
    if is_factor(y):  # classification

        classes, n_elem_classes = np.unique(y, return_counts=True)
        n_classes = len(classes)
        y_as_classes = y.copy()
        freqs_hist = np.zeros_like(n_elem_classes, dtype=float)
        if verbose is True:
            print(f"creating breaks...")
        if n_jobs is None:
            for i in range(len(n_elem_classes)):
                freqs_hist[i] = float(n_elem_classes[i]) / n_obs
        else:

            def get_freqs_hist(i):
                return float(n_elem_classes[i]) / n_obs

            freqs_hist = Parallel(n_jobs=n_jobs)(
                delayed(get_freqs_hist)(i) for i in range(len(n_elem_classes))
            )

    else:  # regression

        h = np.histogram(y, bins="auto")
        n_elem_classes = np.asarray(h[0], dtype=np.int32)
        freqs_hist = np.zeros_like(n_elem_classes, dtype=float)

        if verbose is True:
            print(f"creating breaks...")
        if n_jobs is None:
            for i in range(len(n_elem_classes)):
                freqs_hist[i] = float(n_elem_classes[i]) / n_obs
        else:

            def get_freqs_hist(i):
                return float(n_elem_classes[i]) / n_obs

            freqs_hist = Parallel(n_jobs=n_jobs)(
                delayed(get_freqs_hist)(i) for i in range(len(n_elem_classes))
            )

        breaks = h[1]
        n_breaks_1 = len(breaks) - 1
        classes = range(n_breaks_1)
        n_classes = n_breaks_1
        y_as_classes = np.zeros_like(y, dtype=int)

        for i in classes:
            y_as_classes[(y > breaks[i]) * (y <= breaks[i + 1])] = int(i)

    # main loop ----

    if verbose is True:
        print(f"main loop...")

    if n_jobs is None:

        if verbose is True:
            iterator = tqdm(range(n_classes))
        else:
            iterator = range(n_classes)

        for i in iterator:
            bool_class_i = y_as_classes == classes[i]
            index_class_i = np.asarray(
                np.where(bool_class_i == True)[0], dtype=np.int32
            )
            if np.sum(bool_class_i) > 1:  # at least 2 elements in class  #i
                np.random.seed(seed + i)
                index.extend(
                    np.random.choice(
                        index_class_i,
                        size=int(n_obs_out * freqs_hist[i]),  # output size
                        replace=True,
                    ).tolist()
                )
            else:  # only one element in class
                try:
                    index.append(index_class_i[0])
                except:
                    pass

    else:  # parallel execution

        def get_index(i):
            bool_class_i = y_as_classes == classes[i]
            index_class_i = np.asarray(
                np.where(bool_class_i == True)[0], dtype=np.int32
            )
            if np.sum(bool_class_i) > 1:  # at least 2 elements in class  #i
                np.random.seed(seed + i)
                return np.random.choice(
                    index_class_i,
                    size=int(n_obs_out * freqs_hist[i]),  # output size
                    replace=True,
                ).tolist()
            else:  # only one element in class
                try:
                    return index_class_i[0]
                except:
                    pass

        if verbose is True:
            index = Parallel(n_jobs=n_jobs)(
                delayed(get_index)(i) for i in tqdm(range(n_classes))
            )
        else:
            index = Parallel(n_jobs=n_jobs)(
                delayed(get_index)(i) for i in range(n_classes)
            )

    try:
        return np.asarray(flatten(index))
    except:
        return np.asarray(index)



# Authors: Thierry Moudiki
#
# License: BSD 3

import numpy as np
import sklearn.metrics as skm2
from ..utils import matrixops as mo
from ..utils import misc as mx
from ..base import Base
from copy import deepcopy
from scipy.special import expit
from sklearn.base import ClassifierMixin


class MultitaskClassifier(Base, ClassifierMixin):
    """Multitask Classification model based on regression models, with shared covariates

    Parameters:

        obj: object
            any object (must be a regression model) containing a method fit (obj.fit())
            and a method predict (obj.predict())

        n_hidden_features: int
            number of nodes in the hidden layer

        activation_name: str
            activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'

        a: float
            hyperparameter for 'prelu' or 'elu' activation function

        nodes_sim: str
            type of simulation for the nodes: 'sobol', 'hammersley', 'halton',
            'uniform'

        bias: boolean
            indicates if the hidden layer contains a bias term (True) or not
            (False)

        dropout: float
            regularization parameter; (random) percentage of nodes dropped out
            of the training

        direct_link: boolean
            indicates if the original predictors are included (True) in model's
            fitting or not (False)

        n_clusters: int
            number of clusters for 'kmeans' or 'gmm' clustering (could be 0:
                no clustering)

        cluster_encode: bool
            defines how the variable containing clusters is treated (default is one-hot)
            if `False`, then labels are used, without one-hot encoding

        type_clust: str
            type of clustering method: currently k-means ('kmeans') or Gaussian
            Mixture Model ('gmm')

        type_scaling: a tuple of 3 strings
            scaling methods for inputs, hidden layer, and clustering respectively
            (and when relevant).
            Currently available: standardization ('std') or MinMax scaling ('minmax')

        col_sample: float
            percentage of covariates randomly chosen for training

        row_sample: float
            percentage of rows chosen for training, by stratified bootstrapping

        seed: int
            reproducibility seed for nodes_sim=='uniform'

        backend: str
            "cpu" or "gpu" or "tpu"

    Attributes:

        fit_objs_: dict
            objects adjusted to each individual time series

        n_classes_: int
            number of classes for the classifier

    Examples:

    See also [https://github.com/Techtonique/nnetsauce/blob/master/examples/mtask_classification.py](https://github.com/Techtonique/nnetsauce/blob/master/examples/mtask_classification.py)

    ```python
    import nnetsauce as ns
    import numpy as np
    from sklearn.datasets import load_breast_cancer
    from sklearn.linear_model import LinearRegression
    from sklearn.model_selection import train_test_split
    from sklearn import metrics
    from time import time

    breast_cancer = load_breast_cancer()
    Z = breast_cancer.data
    t = breast_cancer.target

    X_train, X_test, y_train, y_test = train_test_split(Z, t, test_size=0.2,
                                                        random_state=123+2*10)

    # Linear Regression is used
    regr = LinearRegression()
    fit_obj = ns.MultitaskClassifier(regr, n_hidden_features=5,
                                n_clusters=2, type_clust="gmm")

    start = time()
    fit_obj.fit(X_train, y_train)
    print(f"Elapsed {time() - start}")

    print(fit_obj.score(X_test, y_test))
    print(fit_obj.score(X_test, y_test, scoring="roc_auc"))

    start = time()
    preds = fit_obj.predict(X_test)
    print(f"Elapsed {time() - start}")
    print(metrics.classification_report(preds, y_test))
    ```

    """

    # construct the object -----

    def __init__(
        self,
        obj,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        seed=123,
        backend="cpu",
    ):
        super().__init__(
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
            backend=backend,
        )

        self.type_fit = "classification"
        self.obj = obj
        self.fit_objs_ = {}

    def fit(self, X, y, sample_weight=None, **kwargs):
        """Fit MultitaskClassifier to training data (X, y).

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            y: array-like, shape = [n_samples]
                Target values.

            **kwargs: additional parameters to be passed to
                    self.cook_training_set or self.obj.fit

        Returns:

            self: object

        """

        assert mx.is_factor(y), "y must contain only integers"

        output_y, scaled_Z = self.cook_training_set(y=y, X=X, **kwargs)

        self.classes_ = np.unique(y)  # for compatibility with sklearn
        self.n_classes_ = len(self.classes_)  # for compatibility with sklearn

        # multitask response
        Y = mo.one_hot_encode2(output_y, self.n_classes_)

        # if sample_weight is None:
        for i in range(self.n_classes_):
            self.fit_objs_[i] = deepcopy(self.obj.fit(scaled_Z, Y[:, i], **kwargs))

        self.classes_ = np.unique(y)
        return self

    def predict(self, X, **kwargs):
        """Predict test data X.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to
                    self.cook_test_set

        Returns:

            model predictions: {array-like}

        """
        return np.argmax(self.predict_proba(X, **kwargs), axis=1)

    def predict_proba(self, X, **kwargs):
        """Predict probabilities for test data X.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to
                    self.cook_test_set

        Returns:

            probability estimates for test data: {array-like}

        """

        shape_X = X.shape

        probs = np.zeros((shape_X[0], self.n_classes_))

        if len(shape_X) == 1:
            n_features = shape_X[0]

            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

            Z = self.cook_test_set(new_X, **kwargs)

            # loop on all the classes
            for i in range(self.n_classes_):
                probs[:, i] = self.fit_objs_[i].predict(Z, **kwargs)[0]

        else:
            Z = self.cook_test_set(X, **kwargs)

            # loop on all the classes
            for i in range(self.n_classes_):
                probs[:, i] = self.fit_objs_[i].predict(Z, **kwargs)

        expit_raw_probs = expit(probs)

        return expit_raw_probs / expit_raw_probs.sum(axis=1)[:, None]

    def decision_function(self, X, **kwargs):
        """Compute the decision function of X.

        Parameters:
            X: {array-like}, shape = [n_samples, n_features]
                Samples to compute decision function for.

            **kwargs: additional parameters to be passed to
                    self.cook_test_set

        Returns:
            array-like of shape (n_samples,) or (n_samples, n_classes)
            Decision function of the input samples. The order of outputs is the same
            as that of the classes passed to fit.
        """
        if not hasattr(self.obj, "decision_function"):
            # If base classifier doesn't have decision_function, use predict_proba
            proba = self.predict_proba(X, **kwargs)
            if proba.shape[1] == 2:
                return proba[:, 1]  # For binary classification
            return proba  # For multiclass

        if len(X.shape) == 1:
            n_features = X.shape[0]
            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

            return (
                self.obj.decision_function(
                    self.cook_test_set(new_X, **kwargs), **kwargs
                )
            )[0]

        return self.obj.decision_function(self.cook_test_set(X, **kwargs), **kwargs)



import numpy as np
import pandas as pd


def dict_to_dataframe_series(data, series_names):
    df = pd.DataFrame(
        np.zeros((len(data["Model"]), 2)), columns=["Model", "Time Taken"]
    )
    for key, value in data.items():
        if all(hasattr(elt, "__len__") for elt in value) and key not in (
            "Model",
            "Time Taken",
        ):
            for i, elt1 in enumerate(value):
                for j, elt2 in enumerate(elt1):
                    df.loc[i, f"{key}_{series_names[j]}"] = elt2
        else:
            df[key] = value
    return df


# flatten list of lists
# flatten = lambda l: [item for sublist in l for item in sublist]
def flatten(x):
    res = []

    for elt1 in x:
        try:
            for elt2 in elt1:
                res.append(elt2)

        except:
            res.append(elt1)

    return res


# check if the response contains only integers
def is_factor(y):
    """
    Determine if the target variable `y` is for classification (True) or regression (False).
    
    Parameters:
    y : array-like
        Target variable (labels/response variable)
    
    Returns:
    bool
        True if `y` is categorical (classification), False if numeric (regression)
    """
    y_array = np.asarray(y)
    
    # Boolean → classification
    if y_array.dtype == bool:
        return True
    
    # Strings or objects → classification
    if y_array.dtype.kind in ['U', 'S', 'O']:
        return True
    
    # Numeric types (int/float) with few unique values → classification
    if y_array.dtype.kind in ['i', 'u', 'f']:
        unique_values = np.unique(y_array[~np.isnan(y_array)])  # Exclude NaNs
        if len(unique_values) <= 10:  # Threshold for classification
            return True
    
    # Otherwise → regression
    return False

# check if x is float
def is_float(x):
    return isinstance(x, float)


# check if x is int
def is_int(x):
    try:
        return int(x) == x
    except:
        return False


# merge two dictionaries
def merge_two_dicts(x, y):
    z = x.copy()
    z.update(y)
    return z


# apply function to a tuple (element-wise)
def tuple_map(x, foo):
    return tuple(map(foo, x))



from .lmfuncs import beta_hat, inv_penalized_cov
from .matrixops import (
    cbind,
    delete_last_columns,
    rbind,
    convert_df_to_numeric,
    crossprod,
    tcrossprod,
    to_np_array,
)
from .misc import (
    dict_to_dataframe_series,
    merge_two_dicts,
    is_factor,
    tuple_map,
)
from .model_selection import TimeSeriesSplit, cross_val_score
from .progress_bar import Progbar
from .psdcheck import isPD, nearestPD
from .timeseries import (
    compute_output_dates,
    coverage,
    create_lags,
    create_train_inputs,
    reformat_response,
    winkler_score,
    mean_errors,
)


__all__ = [
    "beta_hat",
    "inv_penalized_cov",
    "cbind",
    "delete_last_columns",
    "dict_to_dataframe_series",
    "rbind",
    "convert_df_to_numeric",
    "crossprod",
    "tcrossprod",
    "to_np_array",
    "merge_two_dicts",
    "is_factor",
    "isPD",
    "nearestPD",
    "compute_output_dates",
    "create_lags",
    "create_train_inputs",
    "reformat_response",
    "tuple_map",
    "TimeSeriesSplit",
    "Progbar",
    "coverage",
    "winkler_score",
    "mean_errors",
    "cross_val_score",
]
